from .actor_critic import ActorCriticNet, SharedBodyACNet
from .block import ConvBody, DqnConv, FcBody, ResNetBody, LinearHead, NetworkBlock
from .block import calc_cnn_hidden, make_cnns
from .init import InitFn, Initializer
from .option_critic import OptionCriticNet, SharedBodyOCNet
from .policy import Policy, PolicyDist
from .recurrent import DummyRnn, GruBlock, LstmBlock, RnnBlock, RnnState
from .value import QFunction, QValueNet



from numpy import ndarray
from typing import Callable, Generic, List, NamedTuple, Tuple
from .array_deque import ArrayDeque
from .base import ReplayFeed, ReplayBuffer
from ..envs.ext import State
from ..utils import sample_indices
from ..prelude import GenericNamedMeta


class UniformReplayBuffer(ReplayBuffer, Generic[ReplayFeed]):
    def __init__(self, feed: ReplayFeed, capacity: int = 1000) -> None:
        super().__init__(feed)
        self.buf = ArrayDeque(capacity=capacity)
        self.cap = capacity

    def append(self, *args) -> None:
        exp = self.feed(*args)
        self.buf.push_back(exp)
        if len(self) > self.cap:
            self.buf.pop_front()

    def sample(self, batch_size: int) -> List[ReplayFeed]:
        return [self.buf[idx] for idx in sample_indices(len(self.buf), batch_size)]

    def __len__(self):
        return len(self.buf)


class DqnReplayFeed(NamedTuple, Generic[State], metaclass=GenericNamedMeta):
    state: State
    action: int
    reward: float
    next_state: State
    done: bool

    def to_ndarray(
            self,
            wrap: Callable[[State], ndarray]
    ) -> Tuple[ndarray, int, float, ndarray, bool]:
        return (
            wrap(self.state),
            self.action,
            self.reward,
            wrap(self.next_state),
            self.done,
        )




import numpy as np
from pathlib import Path
from typing import List, Optional, Tuple
from .agents import Agent, EpisodeResult
from .prelude import Array
from .utils.log import ExperimentLog
from .utils.misc import has_freq_in_interval
from .lib.mpi import IS_MPI_ROOT

SAVE_FILE_DEFAULT = 'rainy-agent.pth'
SAVE_FILE_OLD = 'rainy-agent.save'
ACTION_FILE_DEFAULT = 'actions.json'


def eval_impl(
        ag: Agent,
        save_file: Optional[Path],
        render: bool = False,
) -> List[EpisodeResult]:
    n = ag.config.eval_times
    ag.set_mode(train=False)
    if save_file is not None and save_file.is_file():
        res = [ag.eval_and_save(save_file.as_posix(), render=render) for _ in range(n)]
    elif hasattr(ag, 'eval_parallel') and not render:
        res = ag.eval_parallel(n)  # type: ignore
    else:
        res = [ag.eval_episode(render=render) for _ in range(n)]
    ag.set_mode(train=True)
    return res


def _reward_and_length(results: List[EpisodeResult]) -> Tuple[Array[float], Array[float]]:
    rewards = np.array([t.reward for t in results])
    length = np.array([t.length for t in results])
    return rewards, length


def train_agent(
        ag: Agent,
        episode_offset: int = 0,
        saveid_start: int = 0,
        save_file_name: str = SAVE_FILE_DEFAULT,
        action_file_name: str = ACTION_FILE_DEFAULT,
) -> None:
    action_file = Path(action_file_name)

    def log_episode(episodes: int, res: List[EpisodeResult]) -> None:
        rewards, length = _reward_and_length(res)
        ag.logger.exp('train', {
            'episodes': episodes + episode_offset,
            'update-steps': ag.update_steps,
            'reward-mean': float(np.mean(rewards)),
            'reward-min': float(np.min(rewards)),
            'reward-max': float(np.max(rewards)),
            'reward-stdev': float(np.std(rewards)),
            'length-mean': int(np.mean(length)),
        })

    def log_eval() -> None:
        log_dir = ag.logger.log_dir
        if ag.config.save_eval_actions and log_dir:
            fname = log_dir.joinpath('{}-{}{}'.format(
                action_file.stem,
                episodes,
                action_file.suffix
            ))
            res = eval_impl(ag, fname, False)
        else:
            res = eval_impl(ag, None, False)
        rewards, length = _reward_and_length(res)
        ag.logger.exp('eval', {
            'total-steps': ag.total_steps,
            'update-steps': ag.update_steps,
            'reward-mean': float(np.mean(rewards)),
            'reward-min': float(np.min(rewards)),
            'reward-max': float(np.max(rewards)),
            'reward-stdev': float(np.std(rewards)),
            'length-mean': float(np.mean(length)),
        })

    def truncate_episode(episodes: int, freq: Optional[int]) -> int:
        return episodes - episodes % freq if freq else episodes

    episodes = 0
    steps = 0
    save_id = saveid_start
    results: List[EpisodeResult] = []

    for res in ag.train_episodes(ag.config.max_steps):
        ep_len = len(res)
        episodes += ep_len
        step_diff = ag.total_steps - steps
        steps = ag.total_steps
        results += res
        if has_freq_in_interval(episodes, ep_len, ag.config.episode_log_freq) and IS_MPI_ROOT:
            eps = truncate_episode(episodes, ag.config.episode_log_freq)
            log_episode(eps, results[:eps])
            results = results[eps:]
        if has_freq_in_interval(steps, step_diff, ag.config.eval_freq) and IS_MPI_ROOT:
            log_eval()
        if has_freq_in_interval(steps, step_diff, ag.config.save_freq) and IS_MPI_ROOT:
            ag.save(save_file_name + '.{}'.format(save_id))
            save_id += 1
    log_eval()
    ag.save(save_file_name)
    ag.close()


def _load_agent(load_file_name: str, logdir_path: Path, ag: Agent) -> bool:
    def _try_load(fname: str) -> bool:
        p = logdir_path.joinpath(fname)
        if p.exists():
            ag.load(p.as_posix())
            return True
        else:
            return False
    while _try_load(load_file_name) is False:
        if load_file_name == SAVE_FILE_DEFAULT:
            load_file_name = SAVE_FILE_OLD
            continue
        return False
    return True


def retrain_agent(
        ag: Agent,
        log: ExperimentLog,
        load_file_name: str = SAVE_FILE_DEFAULT,
        additional_steps: int = 100
) -> None:
    path = log.log_path.parent
    if not _load_agent(load_file_name, path, ag):
        raise ValueError('Load file {} does not exists'.format(load_file_name))
    episodes, total = 0, 0
    for d in reversed(log.unwrapped):
        if episodes == 0 and 'episodes' in d:
            episodes = d['episodes']
        if total == 0 and 'total-steps' in d:
            total = d['total-steps']
        if episodes > 0 and total > 0:
            break
    save_files = [f for f in path.glob(SAVE_FILE_DEFAULT + '.*')]
    if len(save_files) > 0:
        save_id = len(save_files)
    else:
        save_id = 0
    ag.total_steps = total
    ag.config.max_steps += additional_steps
    train_agent(ag, episode_offset=episodes, saveid_start=save_id)


def eval_agent(
        ag: Agent,
        log_dir: str,
        load_file_name: str = SAVE_FILE_DEFAULT,
        render: bool = False,
        replay: bool = False,
        action_file: Optional[str] = None,
) -> None:
    path = Path(log_dir)
    if not _load_agent(load_file_name, path, ag):
        raise ValueError('Load file {} does not exists'.format())
    if ag.config.save_eval_actions and action_file is not None:
        res = eval_impl(ag, path.joinpath(action_file), render)
    else:
        res = eval_impl(ag, None, render)
    print('{}'.format(res))
    if render:
        input('--Press Enter to exit--')
    if replay:
        try:
            ag.config.eval_env.unwrapped.replay()
        except Exception:
            print('--replay was specified, but environment has no function named replay')
    ag.close()


def random_agent(
        ag: Agent,
        render: bool = False,
        replay: bool = False,
        action_file: Optional[str] = None
) -> None:
    if action_file:
        res = ag.random_and_save(action_file, render=render)
    else:
        res = ag.random_episode(render=render)
    print('{}'.format(res))
    if replay:
        try:
            ag.config.eval_env.unwrapped.replay()
        except Exception:
            print('--replay was specified, but environment has no function named replay')
    ag.close()



from abc import ABC, abstractmethod
import torch
from torch import nn, Tensor
from typing import Generic, Iterable, List, Optional, Sequence, Tuple, TypeVar
from .init import lstm_bias, Initializer
from ..prelude import Index, Self
from ..utils import Device


class RnnState(ABC):
    @abstractmethod
    def __getitem__(self, x: Index) -> Self:
        pass

    @abstractmethod
    def __setitem__(self, x: Index, value: Self) -> None:
        pass

    @abstractmethod
    def fill_(self, f: float) -> None:
        pass

    @abstractmethod
    def unsqueeze(self) -> Self:
        pass


RS = TypeVar('RS', bound=RnnState)


class RnnBlock(Generic[RS], nn.Module):
    def __init__(self, input_dim: int, output_dim: int) -> None:
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim

    @abstractmethod
    def forward(self, x: Tensor, hidden: RS, masks: Optional[Tensor]) -> Tuple[Tensor, RS]:
        pass

    @abstractmethod
    def initial_state(self, batch_size: int, device: Device) -> RS:
        pass


def _apply_mask(mask: Optional[Tensor], *args) -> Sequence[Tensor]:
    if mask is None:
        return tuple(x.unsqueeze(0) for x in args)
    else:
        m = mask.view(1, -1, 1)
        return tuple(x * m for x in args)


@torch.jit.script
def _reshape_batch(x: Tensor, mask: Optional[Tensor], nsteps: int) -> Tuple[Tensor, Tensor]:
    x = x.view(nsteps, -1, x.size(-1))
    if mask is None:
        return x, torch.ones_like(x[:, :, 0])
    else:
        return x, mask.view(nsteps, -1)


def _haszero_iter(mask: Tensor, nstep: int) -> Iterable[Tuple[int, int]]:
    has_zeros = (mask[1:] == 0.0).any(dim=-1).nonzero().squeeze().cpu()
    if has_zeros.dim() == 0:
        haszero = [int(has_zeros.item() + 1)]
    else:
        haszero = (has_zeros + 1).tolist()
    return zip([0] + haszero, haszero + [nstep])


class LstmState(RnnState):
    def __init__(self, h: Tensor, c: Tensor, squeeze: bool = True) -> None:
        self.h = h
        self.c = c
        if squeeze:
            self.h.squeeze_(0)
            self.c.squeeze_(0)

    def __getitem__(self, x: Index) -> Self:
        return LstmState(self.h[x], self.c[x])

    def __setitem__(self, x: Index, value: Self) -> None:
        self.h[x] = value.h[x]
        self.c[x] = value.c[x]

    def fill_(self, f: float) -> None:
        self.h.fill_(f)
        self.c.fill_(f)

    def mul_(self, x: Tensor) -> None:
        self.h.mul_(x)
        self.c.mul_(x)

    def unsqueeze(self) -> Self:
        return LstmState(self.h.unsqueeze(0), self.c.unsqueeze(0))


class LstmBlock(RnnBlock[LstmState]):
    def __init__(
            self,
            input_dim: int,
            output_dim: int,
            initializer: Initializer = Initializer(bias_init=lstm_bias()),
            **kwargs
    ) -> None:
        super().__init__(input_dim, output_dim)
        self.lstm = nn.LSTM(input_dim, output_dim, **kwargs)
        initializer(self.lstm)

    def forward(
            self,
            x: Tensor,
            hidden: LstmState,
            mask_: Optional[Tensor] = None
    ) -> Tuple[Tensor, LstmState]:
        in_shape = x.shape
        if in_shape == hidden.h.shape:
            out, (h, c) = self.lstm(x.unsqueeze(0), _apply_mask(mask_, hidden.h, hidden.c))
            return out.squeeze(0), LstmState(h, c)
        # forward Nsteps altogether
        nsteps = in_shape[0] // hidden.h.size(0)
        x, mask = _reshape_batch(x, mask_, nsteps)
        res, h, c = [], hidden.h, hidden.c
        for start, end in _haszero_iter(mask, nsteps):
            m = mask[start].view(1, -1, 1)
            processed, (h, c) = self.lstm(x[start:end], (h * m, c * m))
            res.append(processed)
        return torch.cat(res).view(in_shape), LstmState(h, c)

    def initial_state(self, batch_size: int, device: Device) -> LstmState:
        zeros = device.zeros((batch_size, self.input_dim))
        return LstmState(zeros, zeros, squeeze=False)


class GruState(RnnState):
    def __init__(self, h: Tensor) -> None:
        self.h = h

    def __getitem__(self, x: Index) -> Self:
        return GruState(self.h[x])

    def __setitem__(self, x: Index, value: Self) -> None:
        self.h[x] = value.h[x]

    def fill_(self, f: float) -> None:
        self.h.fill_(f)

    def mul_(self, x: Tensor) -> None:
        self.h.mul_(x)

    def unsqueeze(self) -> Self:
        return GruState(self.h.unsqueeze(0))


class GruBlock(RnnBlock[GruState]):
    def __init__(
            self,
            input_dim: int,
            output_dim: int,
            initializer: Initializer = Initializer(),
            **kwargs
    ) -> None:
        super().__init__(input_dim, output_dim)
        self.gru = nn.GRU(input_dim, output_dim, **kwargs)
        initializer(self.gru)

    def forward(
            self,
            x: Tensor,
            hidden: GruState,
            mask_: Optional[Tensor] = None
    ) -> Tuple[Tensor, GruState]:
        in_shape = x.shape
        if in_shape == hidden.h.shape:
            out, h = self.gru(x.unsqueeze(0), *_apply_mask(mask_, hidden.h))
            return out.squeeze(0), GruState(h.squeeze_(0))
        # forward Nsteps altogether
        nsteps = in_shape[0] // hidden.h.size(0)
        x, mask = _reshape_batch(x, mask_, nsteps)
        res, h = [], hidden.h
        for start, end in _haszero_iter(mask, nsteps):
            processed, h = self.gru(x[start:end], h * mask[start].view(1, -1, 1))
            res.append(processed)
        return torch.cat(res).view(in_shape), GruState(h.squeeze_(0))

    def initial_state(self, batch_size: int, device: Device) -> GruState:
        return GruState(device.zeros((batch_size, self.input_dim)))


class DummyState(RnnState):
    def __getitem__(self, x: Index) -> Self:
        return self

    def __setitem__(self, x: Index, value: Self) -> None:
        pass

    def fill_(self, f: float) -> None:
        pass

    def unsqueeze(self) -> Self:
        return self


class DummyRnn(RnnBlock[DummyState]):
    DUMMY_STATE = DummyState()

    def __init__(self, *args, **kwargs) -> None:
        nn.Module.__init__(self)

    def forward(
            self,
            x: Tensor,
            hidden: DummyState = DUMMY_STATE,
            mask: Optional[Tensor] = None,
    ) -> Tuple[Tensor, DummyState]:
        return x, hidden

    def initial_state(self, batch_size: int, device: Device) -> DummyState:
        return self.DUMMY_STATE



from abc import ABC, abstractmethod
import numpy as np
from numpy import ndarray
from torch import nn, Tensor
from typing import Tuple, Union
from .block import DqnConv, FcBody, LinearHead, NetworkBlock
from .prelude import NetFn
from ..utils import Device


class QFunction(ABC):
    @abstractmethod
    def action_values(self, state: ndarray, nostack: bool = False) -> Tensor:
        pass

    @property
    @abstractmethod
    def state_dim(self) -> Tuple[int, ...]:
        pass

    @property
    @abstractmethod
    def action_dim(self) -> int:
        pass


class QValueNet(QFunction, nn.Module):
    """State -> [Value..]
    """
    def __init__(self, body: NetworkBlock, head: NetworkBlock, device: Device = Device()) -> None:
        assert body.output_dim == np.prod(head.input_dim), \
            'body output and head input must have a same dimention'
        super().__init__()
        self.head = head
        self.body = body
        if device.is_multi_gpu():
            self.body = device.data_parallel(body)  # type: ignore
        self.device = device
        self.to(self.device.unwrapped)

    def action_values(self, state: ndarray, nostack: bool = False) -> Tensor:
        if nostack:
            return self.forward(state)
        else:
            return self.forward(np.stack([state]))

    def forward(self, x: Union[ndarray, Tensor]) -> Tensor:
        x = self.device.tensor(x)
        x = self.body(x)
        x = self.head(x)
        return x

    @property
    def state_dim(self) -> Tuple[int, ...]:
        return self.body.input_dim

    @property
    def action_dim(self) -> int:
        return self.head.output_dim


def dqn_conv(*args, **kwargs) -> NetFn:
    def _net(state_dim: Tuple[int, int, int], action_dim: int, device: Device) -> QValueNet:
        body = DqnConv(state_dim, *args, **kwargs)
        head = LinearHead(body.output_dim, action_dim)
        return QValueNet(body, head, device=device)
    return _net  # type: ignore


def fc(*args, **kwargs) -> NetFn:
    def _net(state_dim: Tuple[int, ...], action_dim: int, device: Device) -> QValueNet:
        body = FcBody(state_dim[0], *args, **kwargs)
        head = LinearHead(body.output_dim, action_dim)
        return QValueNet(body, head, device=device)
    return _net



import numpy as np
import torch
from torch import ByteTensor, LongTensor, nn, Tensor
from typing import List, Optional, Tuple
from .base import NStepParallelAgent
from ..config import Config
from ..lib.explore import EpsGreedy
from ..lib.rollout import RolloutStorage
from ..net import OptionCriticNet
from ..net.policy import Policy, BernoulliPolicy
from ..prelude import Action, Array, State
from ..utils import Device


class AocRolloutStorage(RolloutStorage[State]):
    def __init__(self, nsteps: int, nworkers: int, device: Device, num_options: int) -> None:
        super().__init__(nsteps, nworkers, device)
        self.options = [self.device.zeros(self.nworkers, dtype=torch.long)]
        self.is_new_options = [self.device.ones(self.nworkers, dtype=torch.uint8)]
        self.epsilons: List[float] = []
        self.beta_adv = torch.zeros_like(self.batch_values)
        self.noptions = num_options
        self.worker_indices = self.device.indices(self.nworkers)

    def reset(self) -> None:
        super().reset()
        self.options = [self.options[-1]]
        self.is_new_options = [self.is_new_options[-1]]
        self.epsilons.clear()

    def push_options(self, option: LongTensor, is_new_option: Tensor, epsilon: float) -> None:
        self.options.append(option)
        self.is_new_options.append(is_new_option)
        self.epsilons.append(epsilon)

    def batch_options(self) -> Tuple[Tensor, Tensor]:
        batched = torch.cat(self.options, dim=0)
        return batched[:-self.nworkers], batched[self.nworkers:]

    def calc_returns(self, next_value: Tensor, gamma: float, delib_cost: float) -> None:
        self.returns[-1] = next_value
        rewards = self.device.tensor(self.rewards)
        for i in reversed(range(self.nsteps)):
            ret = gamma * self.masks[i + 1] * self.returns[i + 1] + rewards[i]
            self.returns[i] = ret - self.is_new_options[i].float() * self.masks[i] * delib_cost
            opt = self.options[i + 1]
            opt_q, eps = self.values[i], self.epsilons[i]
            self.advs[i] = self.returns[i] - opt_q[self.worker_indices, opt]
            v = (1 - eps) * opt_q.max(dim=-1)[0] + eps * opt_q.mean(dim=-1)
            self.beta_adv[i] = opt_q[self.worker_indices, opt].mul(v)


class AocAgent(NStepParallelAgent[State]):
    """AOC: Adavantage Option Critic
    It's a synchronous batched version of A2OC: Asynchronou Adavantage Option Critic
    """
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.net: OptionCriticNet = config.net('option-critic')  # type: ignore
        self.noptions = self.net.num_options
        self.optimizer = config.optimizer(self.net.parameters())
        self.worker_indices = config.device.indices(config.nworkers)
        self.batch_indices = config.device.indices(config.nworkers * config.nsteps)
        self.storage: AocRolloutStorage[State] = \
            AocRolloutStorage(config.nsteps, config.nworkers, config.device, self.noptions)
        self.opt_explorer: EpsGreedy = config.explorer()  # type: ignore
        if not isinstance(self.opt_explorer, EpsGreedy):
            return ValueError('Currently only Epsilon Greedy is supported as Explorer')
        self.eval_prev_options: LongTensor = config.device.zeros(config.nworkers, dtype=torch.long)

    def members_to_save(self) -> Tuple[str, ...]:
        return 'net', 'opt_explorer'

    def _reset(self, initial_states: Array[State]) -> None:
        self.storage.set_initial_state(initial_states)

    def eval_reset(self) -> None:
        self.eval_prev_options.fill_(0)

    def sample_options(
            self,
            opt_q: Tensor,
            beta: BernoulliPolicy,
            prev_options: LongTensor,
    ) -> Tuple[LongTensor, ByteTensor]:
        current_beta = beta[self.worker_indices, prev_options]
        do_options_end = current_beta.action()
        use_new_options = do_options_end.add(1.0 - self.storage.masks[-1]) > 0.9
        epsgreedy_options = self.opt_explorer.select_from_value(opt_q)
        options = torch.where(use_new_options, epsgreedy_options, prev_options)
        return options, use_new_options  # type: ignore

    @torch.no_grad()
    def _eval_policy(self, states: Array, indices: Tensor) -> Policy:
        opt_policy, opt_q, beta = self.net(states)
        options, _ = self.sample_options(opt_q, beta, self.eval_prev_options)
        self.eval_prev_options = options
        return opt_policy[indices, options]

    def eval_action(self, state: Array) -> Action:
        if len(state.shape) == len(self.net.state_dim):
            # treat as batch_size == nworkers
            state = np.stack([state] * self.config.nworkers)
        policy = self._eval_policy(state, self.config.device.tensor([0], dtype=torch.long))[0]
        act = policy.best_action() if self.config.eval_deterministic else policy.action()
        return act.squeeze().cpu().numpy()

    def eval_action_parallel(
            self,
            states: Array,
            mask: torch.Tensor,
            ent: Optional[Array[float]] = None
    ) -> Array[Action]:
        policy = self._eval_policy(states, self.worker_indices)
        if ent is not None:
            ent += policy.entropy().cpu().numpy()
        act = policy.best_action() if self.config.eval_deterministic else policy.action()
        return act.squeeze().cpu().numpy()

    @property
    def prev_options(self) -> LongTensor:
        return self.storage.options[-1]  # type: ignore

    @property
    def prev_is_new_options(self) -> ByteTensor:
        return self.storage.is_new_options[-1]  # type: ignore

    @torch.no_grad()
    def _one_step(self, states: Array[State]) -> Array[State]:
        opt_policy, opt_q, beta = self.net(self.penv.extract(states))
        options, is_new_options = self.sample_options(opt_q, beta, self.prev_options)
        policy = opt_policy[self.worker_indices, options]
        actions = policy.action().squeeze().cpu().numpy()
        next_states, rewards, done, info = self.penv.step(actions)
        self.episode_length += 1
        self.rewards += rewards
        self.report_reward(done, info)
        self.storage.push(next_states, rewards, done, policy=policy, value=opt_q)
        self.storage.push_options(options, is_new_options, self.opt_explorer.epsilon)
        return next_states

    @torch.no_grad()
    def _next_value(self, states: Array[State]) -> Tensor:
        opt_q = self.net.opt_q(self.penv.extract(states))
        current_opt_q = opt_q[self.worker_indices, self.prev_options]
        eps = self.opt_explorer.epsilon
        next_opt_q = (1 - eps) * opt_q.max(dim=-1)[0] + eps * opt_q.mean(-1)
        return torch.where(self.prev_is_new_options, next_opt_q, current_opt_q)

    def nstep(self, states: Array[State]) -> Array[State]:
        for _ in range(self.config.nsteps):
            states = self._one_step(states)

        next_value = self._next_value(states)
        self.storage.calc_returns(
            next_value,
            self.config.discount_factor,
            self.config.opt_delib_cost,
        )

        prev_options, options = self.storage.batch_options()
        adv = self.storage.advs[:-1].flatten()
        beta_adv = self.storage.beta_adv.flatten()
        ret = self.storage.returns[:-1].flatten()
        masks = self.storage.batch_masks()

        opt_policy, opt_q, beta = self.net(self.storage.batch_states(self.penv))
        policy = opt_policy[self.batch_indices, prev_options]
        batch_actions = self.storage.batch_actions()
        policy.set_action(batch_actions)

        policy_loss = -(policy.log_prob() * adv).mean()
        term_prob = beta[self.batch_indices, prev_options].dist.probs
        beta_adv += self.config.opt_delib_cost + self.config.opt_beta_adv_merginal
        beta_loss = term_prob.mul(masks).mul(beta_adv).mean()
        value_loss = (opt_q[self.batch_indices, options] - ret).pow(2).mean()
        entropy = policy.entropy().mean()

        self.optimizer.zero_grad()
        (policy_loss
         + beta_loss
         + self.config.value_loss_weight * 0.5 * value_loss
         - self.config.entropy_weight * entropy).backward()
        nn.utils.clip_grad_norm_(self.net.parameters(), self.config.grad_clip)
        self.optimizer.step()

        self.report_loss(
            policy_loss=policy_loss.item(),
            value_loss=value_loss.item(),
            beta_loss=beta_loss.item(),
            entropy=entropy.item(),
        )
        self.storage.reset()
        return states



from torch import Tensor
from .dqn import DqnAgent
from ..prelude import Array


class DoubleDqnAgent(DqnAgent):
    def _q_next(self, next_states: Array) -> Tensor:
        """Returns Q values of next_states, supposing torch.no_grad() is called
        """
        q_next = self.target_net(next_states)
        action_values = self.net.action_values(next_states, nostack=True)
        return q_next[self.batch_indices, action_values.argmax(dim=-1)]



import numpy as np
import torch
from torch import nn
from typing import Optional, Tuple
from .base import NStepParallelAgent
from ..config import Config
from ..lib.rollout import RolloutStorage
from ..net import ActorCriticNet, Policy, RnnState
from ..prelude import Action, Array, State


class A2cAgent(NStepParallelAgent[State]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.storage: RolloutStorage[State] = \
            RolloutStorage(config.nsteps, config.nworkers, config.device)
        self.net: ActorCriticNet = config.net('actor-critic')  # type: ignore
        self.optimizer = config.optimizer(self.net.parameters())
        self.lr_cooler = config.lr_cooler(self.optimizer.param_groups[0]['lr'])
        self.eval_rnns = self.rnn_init()

    def members_to_save(self) -> Tuple[str, ...]:
        return 'net', 'lr_cooler'

    def set_mode(self, train: bool = True) -> None:
        self.net.train(mode=train)

    def rnn_init(self) -> RnnState:
        return self.net.recurrent_body.initial_state(self.config.nworkers, self.config.device)

    def _reset(self, initial_states: Array[State]) -> None:
        self.storage.set_initial_state(initial_states, self.rnn_init())

    def eval_reset(self) -> None:
        self.eval_rnns.fill_(0.0)

    def eval_action(self, state: Array) -> Action:
        if len(state.shape) == len(self.net.state_dim):
            # treat as batch_size == 1
            state = np.stack([state])
        with torch.no_grad():
            policy, self.eval_rnns[0] = self.net.policy(state, self.eval_rnns[0].unsqueeze())
        if self.config.eval_deterministic:
            return policy.best_action().squeeze().cpu().numpy()
        else:
            return policy.action().squeeze().cpu().numpy()

    def eval_action_parallel(
            self,
            states: Array,
            mask: torch.Tensor,
            ent: Optional[Array[float]] = None
    ) -> Array[Action]:
        with torch.no_grad():
            policy, self.eval_rnns = self.net.policy(states, self.eval_rnns)
        if ent is not None:
            ent += policy.entropy().cpu().numpy()
        if self.config.eval_deterministic:
            return policy.best_action().squeeze().cpu().numpy()
        else:
            return policy.action().squeeze().cpu().numpy()

    def _network_in(self, states: Array[State]) -> Tuple[Array, RnnState, torch.Tensor]:
        return self.penv.extract(states), self.storage.rnn_states[-1], self.storage.masks[-1]

    def _one_step(self, states: Array[State]) -> Array[State]:
        with torch.no_grad():
            policy, value, rnns = self.net(*self._network_in(states))
        next_states, rewards, done, info = self.penv.step(policy.action().squeeze().cpu().numpy())
        self.episode_length += 1
        self.rewards += rewards
        self.report_reward(done, info)
        self.storage.push(next_states, rewards, done, rnn_state=rnns, policy=policy, value=value)
        return next_states

    def _pre_backward(self, _policy: Policy, _value: torch.Tensor) -> None:
        pass

    def _step_optimizer(self) -> None:
        nn.utils.clip_grad_norm_(self.net.parameters(), self.config.grad_clip)
        self.optimizer.step()
        self.lr_cooler.lr_decay(self.optimizer)

    def nstep(self, states: Array[State]) -> Array[State]:
        for _ in range(self.config.nsteps):
            states = self._one_step(states)
        with torch.no_grad():
            next_value = self.net.value(*self._network_in(states))
        if self.config.use_gae:
            gamma, lambda_ = self.config.discount_factor, self.config.gae_lambda
            self.storage.calc_gae_returns(next_value, gamma, lambda_)
        else:
            self.storage.calc_ac_returns(next_value, self.config.discount_factor)
        policy, value, _ = self.net(
            self.storage.batch_states(self.penv),
            self.storage.rnn_states[0],
            self.storage.batch_masks(),
        )
        policy.set_action(self.storage.batch_actions())

        advantage = self.storage.returns[:-1].flatten() - value
        policy_loss = -(policy.log_prob() * advantage.detach()).mean()
        value_loss = advantage.pow(2).mean()
        entropy_loss = policy.entropy().mean()
        self._pre_backward(policy, value)
        self.optimizer.zero_grad()
        (policy_loss
         + self.config.value_loss_weight * 0.5 * value_loss
         - self.config.entropy_weight * entropy_loss).backward()
        self._step_optimizer()
        self.report_loss(
            policy_loss=policy_loss.item(),
            value_loss=value_loss.item(),
            entropy_loss=entropy_loss.item(),
        )
        self.storage.reset()
        return states



import numpy as np
from torch.utils.data.sampler import BatchSampler, Sampler, SubsetRandomSampler
from typing import Iterator
from ..prelude import Array


def sample_indices(n: int, k: int) -> np.ndarray:
    """Sample k numbers from [0, n)
       Based on https://github.com/chainer/chainerrl/blob/master/chainerrl/misc/random.py
    """
    if 3 * k >= n:
        return np.random.choice(n, k, replace=False)
    else:
        selected = np.repeat(False, n)
        rands = np.random.randint(0, n, size=k * 2)
        j = k
        for i in range(k):
            x = rands[i]
            while selected[x]:
                if j == 2 * k:
                    rands[k:] = np.random.randint(0, n, size=k)
                    j = k
                x = rands[i] = rands[j]
                j += 1
            selected[x] = True
        return rands[:k]


class FeedForwardBatchSampler(BatchSampler):
    def __init__(self, nsteps: int, nworkers: int, batch_size: int) -> None:
        super().__init__(
            SubsetRandomSampler(range(nsteps * nworkers)),
            batch_size=batch_size,
            drop_last=True
        )


class RecurrentBatchSampler(Sampler):
    def __init__(self, nsteps: int, nworkers: int, batch_size: int) -> None:
        if batch_size % nsteps > 0:
            raise ValueError('batch_size must be a multiple of nsteps')
        self.nsteps = nsteps
        self.nworkers = nworkers
        self.batch_size = batch_size

    def __iter__(self) -> Iterator[Array[int]]:
        env_num = self.batch_size // self.nsteps
        total, step = self.nsteps * self.nworkers, self.nworkers
        perm = np.random.permutation(self.nworkers)
        for end in np.arange(env_num, self.nworkers + 1, env_num):
            workers = perm[end - env_num: end]
            batches = np.stack([np.arange(w, total, step) for w in workers], axis=1)
            yield batches.flatten()

    def __len__(self) -> int:
        return (self.nsteps * self.nworkers) // self.batch_size




from collections import defaultdict
from datetime import datetime
import json
import logging
import numpy as np
import git
from pathlib import Path
import sys
from typing import Any, Dict, List, Optional, Set, Union


NORMAL_FORMATTER = logging.Formatter('%(levelname)s %(asctime)s: %(name)s: %(message)s')
JSON_FORMATTER = logging.Formatter('%(levelname)s::%(message)s')
FINGERPRINT = 'fingerprint.txt'
LOGFILE = 'log.txt'
EXP = 5
logging.addLevelName(EXP, 'EXP')


def _load_log_file(file_path: Path) -> List[Dict[str, Any]]:
    with open(file_path.as_posix()) as f:
        lines = f.readlines()
    log = []
    for line in filter(lambda s: s.startswith('EXP::'), lines):
        log.append(json.loads(line[5:]))
    return log


class LogWrapper:
    """Wrapper of filterd log.
    """
    def __init__(
            self,
            name: str,
            inner: List[Dict[str, Any]],
            path: Optional[Path] = None,
    ) -> None:
        self.name = name
        self.inner = inner
        self._available_keys: Set[str] = set()
        self._path = path

    @property
    def unwrapped(self) -> List[Dict[str, Any]]:
        return self.inner

    def keys(self) -> Set[str]:
        if not self._available_keys:
            for log in self.inner:
                for key in log:
                    self._available_keys.add(key)
        return self._available_keys

    def get(self, key: str) -> List[Any]:
        if key not in self.inner[0]:
            raise KeyError(
                'LogWrapper({}) doesn\'t have the logging key {}. Available keys: {}'
                .format(self.name, key, self.keys())
            )
        return [d[key] for d in self.inner]

    def is_empty(self) -> bool:
        return len(self.inner) == 0

    def __repr__(self) -> str:
        return 'LogWrapper({}, {})'.format(self._path, self.name)

    def __getitem__(self, key: str) -> List[Any]:
        return self.get(key)


class ExperimentLog:
    """Structured log file.
       Used to get graphs or else from rainy log files.
    """
    def __init__(self, file_or_dir_name: Union[str, Path], empty: bool = False) -> None:
        path = Path(file_or_dir_name)
        if path.is_dir() and not empty:
            log_path = path.joinpath(LOGFILE)
            self.fingerprint = path.joinpath(FINGERPRINT).read_text()
        else:
            log_path = path
            self.fingerprint = ''
        self.log = [] if empty else _load_log_file(log_path)
        self._available_keys: Set[str] = set()
        self.log_path = log_path

    @property
    def unwrapped(self) -> List[Dict[str, Any]]:
        return self.log

    def keys(self) -> Set[str]:
        if not self._available_keys:
            for log in self.log:
                self._available_keys.add(log['name'])
        return self._available_keys

    def get(self, key: str) -> LogWrapper:
        log = LogWrapper(
            key,
            list(filter(lambda log: log['name'] == key, self.log)),
            self.log_path
        )
        if log.is_empty():
            raise KeyError(
                '{} doesn\'t have the key {}. Available keys: {}'
                .format(self, key, self.keys())
            )
        return log

    def plot_reward(self, batch_size: int, max_steps: int = int(2e7), title: str = '') -> None:
        try:
            import matplotlib.pyplot as plt
        except ModuleNotFoundError as e:
            print('plot_reward need matplotlib installed')
            raise e
        tlog = self.get('train')
        x, y = 'update-steps', 'reward-mean'
        plt.plot(np.array(tlog[x]) * batch_size, tlog[y])
        tick_fractions = np.array([0.1, 0.2, 0.5, 1.0])
        ticks = tick_fractions * max_steps
        MILLION = int(1e6)
        if max_steps >= MILLION:
            tick_names = ["{}M".format(int(tick / 1e6)) for tick in ticks]
        else:
            tick_names = ["{}".format(int(tick)) for tick in ticks]
        plt.xticks(ticks, tick_names)
        plt.title(title)
        plt.xlabel('Frames used for training')
        plt.ylabel(y)
        plt.show()

    def __getitem__(self, key: str) -> LogWrapper:
        return self.get(key)

    def __repr__(self) -> str:
        return 'ExperimentLog({})'.format(self.log_path.as_posix())


class ExpStats:
    """Statictics of loss
    """
    def __init__(self) -> None:
        self.inner: Dict[str, List[float]] = defaultdict(list)

    def update(self, d: Dict[str, float]) -> None:
        for key in d.keys():
            self.inner[key].append(d[key])

    def report_and_reset(self) -> Dict[str, float]:
        res = {}
        for k, v in self.inner.items():
            res[k] = np.array(v).mean()
            v.clear()
        return res


class Logger(logging.Logger):
    ELAPSED = 'elapsed-time'

    def __init__(self) -> None:
        super().__init__('rainy', EXP)

        self._log_dir: Optional[Path] = None
        self.exp_start = datetime.now()
        self.time_offset = 0.0

    def set_dir_from_script_path(
            self,
            script_path_: str,
            comment: Optional[str] = None,
            prefix: str = '',
    ) -> None:
        script_path = Path(script_path_)
        log_dir = script_path.stem + '-' + self.exp_start.strftime("%y%m%d-%H%M%S")
        if prefix:
            log_dir = prefix + '/' + log_dir
        try:
            repo = git.Repo(script_path, search_parent_directories=True)
            head = repo.head.commit
            log_dir += '-' + head.hexsha[:8]
        finally:
            pass
        log_dir_path = Path(log_dir)
        if not log_dir_path.exists():
            log_dir_path.mkdir()
        self.set_dir(log_dir_path, comment=comment)

    @staticmethod
    def filehandler(log_path: Path, level: int) -> logging.Handler:
        if not log_path.exists():
            log_path.touch()
        handler = logging.FileHandler(log_path.as_posix())
        handler.setFormatter(JSON_FORMATTER)
        handler.setLevel(level)
        return handler

    def set_dir(self, log_dir: Path, comment: Optional[str] = None) -> None:
        self._log_dir = log_dir
        finger = log_dir.joinpath(FINGERPRINT)
        with open(finger.as_posix(), 'w') as f:
            f.write('{}\n'.format(self.exp_start))
            if comment:
                f.write(comment)
        handler = self.filehandler(Path(log_dir).joinpath(LOGFILE), EXP)
        self.addHandler(handler)

    def retrive(self, log_dir: str) -> ExperimentLog:
        self._log_dir = Path(log_dir)
        if not self._log_dir.exists():
            raise ValueError('Specified log directory {} does not exist'.format(log_dir))
        logfile = self._log_dir.joinpath(LOGFILE)
        if not logfile.exists():
            raise ValueError('Specified log directory {} does have {}'.format(log_dir, LOGFILE))
        self.addHandler(self.filehandler(logfile, EXP))
        log = ExperimentLog(logfile)
        d = next(filter(lambda d: self.ELAPSED in d, reversed(log.unwrapped)))
        for d in reversed(log.unwrapped):
            if self.ELAPSED in d:
                self.time_offset = d[self.ELAPSED]
                break
        return log

    def set_stderr(self, level: int = EXP) -> None:
        handler = logging.StreamHandler(stream=sys.stderr)
        handler.setFormatter(NORMAL_FORMATTER)
        handler.setLevel(level)
        self.addHandler(handler)

    @property
    def log_dir(self) -> Optional[Path]:
        return self._log_dir

    def exp(self, name: str, msg: dict, *args, **kwargs) -> None:
        """
        For experiment logging. Only dict is enabled as argument
        """
        if not self.isEnabledFor(EXP):
            return
        delta = datetime.now() - self.exp_start
        msg[self.ELAPSED] = delta.total_seconds() + self.time_offset
        msg['name'] = name
        self._log(EXP, json.dumps(msg, sort_keys=True), args, **kwargs)  # type: ignore


class DummyLogger(Logger):
    def __init__(self) -> None:
        pass

    def set_dir_from_script_path(
            self,
            script_path_: str,
            comment: Optional[str] = None,
            prefix: str = '',
    ) -> None:
        pass

    def set_dir(self, log_dir: Path, comment: Optional[str] = None) -> None:
        pass

    def retrive(self, log_dir: str) -> ExperimentLog:
        return ExperimentLog(log_dir, empty=True)

    def set_stderr(self, level: int = EXP) -> None:
        pass

    @property
    def log_dir(self) -> Optional[Path]:
        return None

    def exp(self, name: str, msg: dict, *args, **kwargs) -> None:
        pass



import torch
from torch import Tensor
from torch.optim import Optimizer
from typing import Tuple
from ..prelude import Array
try:
    import horovod.torch as hvd
    hvd.init()
    IS_MPI_ROOT = hvd.rank() == 0

    if hvd.size() == 1:
        raise ModuleNotFoundError('When hvd.size() == 1 we do not use horovod')

    def setup_models(*args) -> None:
        for model in args:
            hvd.broadcast_parameters(model.state_dict(), root_rank=0)

    def setup_optimizer(opt: Optimizer) -> Optimizer:
        hvd.broadcast_optimizer_state(opt, root_rank=0)
        return hvd.DistributedOptimizer(opt)

    def clip_and_step(model: torch.nn.Module, max_norm: float, opt: Optimizer) -> None:
        opt.synchronize()
        torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm)
        with opt.skip_synchronize():
            opt.step()

    def broadcast_model(model: torch.nn.Module) -> None:
        hvd.broadcast_parameters(model.state_dict(), root_rank=0)

    def local_size_and_rank() -> Tuple[int, int]:
        return hvd.local_size(), hvd.local_rank()

    def global_size() -> int:
        return hvd.size()

    def array_reduce_(arr: Array, average: bool = True) -> None:
        t = torch.from_numpy(arr)
        hvd.allreduce_(t, average=average)

    def array_mean_and_var(arr: Array) -> Tuple[Array, Array]:
        m, v = tensor_mean_and_var(torch.from_numpy(arr))
        return m.numpy(), v.numpy()

    def tensor_mean_and_var(t: Tensor) -> Tuple[Tensor, Tensor]:
        mean = hvd.allreduce_(t.mean(dim=0))
        var = hvd.allreduce_((t - mean).pow(2).mean(dim=0))
        return mean, var

except ModuleNotFoundError:

    IS_MPI_ROOT = True

    def setup_models(*args) -> None:
        pass

    def setup_optimizer(opt: Optimizer) -> Optimizer:
        return opt

    def broadcast_model(model: torch.nn.Module) -> None:
        pass

    def local_size_and_rank() -> Tuple[int, int]:
        return 1, 0

    def clip_and_step(model: torch.nn.Module, max_norm: float, opt: Optimizer) -> None:
        torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm)
        opt.step()

    def global_size() -> int:
        return 1

    def array_reduce_(arr: Array, average: bool = True) -> None:
        pass

    def array_mean_and_var(arr: Array) -> Tuple[Array, Array]:
        return arr.mean(axis=0), arr.var(axis=0)

    def tensor_mean_and_var(t: Tensor) -> Tuple[Tensor, Tensor]:
        return t.mean(dim=0), t.var(dim=0, unbiased=False)



"""From https://github.com/openai/baselines/blob/master/baselines/common/running_mean_std.py
"""
import numpy as np
import torch
from typing import Tuple
from ..lib import mpi
from ..prelude import Array
from ..utils import Device
from ..utils.state_dict import TensorStateDict


class RunningMeanStd:
    """Calcurate running mean and variance
    """
    def __init__(self, shape: Tuple[int, ...], epsilon: float = 1.0e-4) -> None:
        self.mean = np.zeros(shape, 'float64')
        self.var = np.ones(shape, 'float64')
        self.count = epsilon

    def update(self, x: Array[float]) -> None:
        x_mean, x_var = mpi.array_mean_and_var(x)
        self.mean, self.var, self.count = _update_rms(
            self.mean,
            self.var,
            self.count,
            x_mean,
            x_var,
            x.shape[0] * mpi.global_size()
        )

    def std(self, eps: float = 1.0e-8) -> Array[float]:
        return np.sqrt(self.var + eps)


def _update_rms(mean, var, count, batch_mean, batch_var, batch_count):
    """https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
    """
    delta = batch_mean - mean
    tot_count = count + batch_count
    new_mean = mean + delta * batch_count / tot_count
    m_a = var * count
    m_b = batch_var * batch_count
    M2 = m_a + m_b + np.square(delta) * count * batch_count / tot_count
    new_var = M2 / tot_count
    new_count = tot_count
    return new_mean, new_var, new_count


class RunningMeanStdTorch(TensorStateDict):
    """Same as RunningMeanStd, but uses PyTorch Tensor
    """
    def __init__(self, shape: torch.Size, device: Device, epsilon: float = 1.0e-4) -> None:
        self.mean = device.zeros(shape, dtype=torch.float64)
        self.var = device.ones(shape, dtype=torch.float64)
        self.count = torch.tensor(epsilon, dtype=torch.float64, device=device.unwrapped)
        self.device = device

    @torch.no_grad()
    def update(self, x: torch.Tensor) -> None:
        x_mean, x_var = mpi.tensor_mean_and_var(x)
        _update_rms_torch(
            self.mean,
            self.var,
            self.count,
            x_mean,
            x_var,
            torch.tensor(x.size(0) * mpi.global_size(), device=self.device.unwrapped)
        )

    def std(self, eps: float = 1.0e-8) -> torch.Tensor:
        return torch.sqrt(self.var + eps)


@torch.jit.script
def _update_rms_torch(mean, var, count, batch_mean, batch_var, batch_count):
    """PyTorch version of _update_rms
    """
    delta = batch_mean - mean
    tot_count = count + batch_count
    mean.add_(delta * batch_count / tot_count)
    m_b = batch_var * batch_count
    delta.pow_(2).mul_(count).mul_(batch_count).div_(tot_count)
    var.mul_(count).add_(m_b).add_(delta).div_(tot_count)
    count.add_(batch_count)



import torch
from torch import Tensor
from typing import Optional


@torch.jit.script
def normalize_(t: Tensor, eps: float) -> None:
    mean = t.mean()
    std = t.std()
    t.sub_(mean).div_(std + eps)


def has_freq_in_interval(turn: int, width: int, freq: Optional[int]) -> bool:
    return freq and turn != 0 and turn // freq != (turn - width) // freq  # type: ignore



import click
from typing import Callable, Optional, Tuple

from .log import ExperimentLog
from ..agents import Agent
from ..config import Config
from ..lib import mpi
from .. import run


@click.group()
@click.option('--gpu', required=False, type=int, help='How many gpus you allow Rainy to use')
@click.option('--envname', type=str, default=None, help='Name of environment passed to config_gen')
@click.option('--seed', type=int, default=None,
              help='Random seed set before training. Left for backward comaptibility')
@click.option('--override', type=str, default='', help='Override string(See README for detail)')
@click.pass_context
def rainy_cli(
        ctx: dict,
        gpu: Tuple[int],
        envname: Optional[str],
        seed: Optional[int],
        override: str
) -> None:
    ctx.obj['gpu'] = gpu
    cfg_gen = ctx.obj['config_gen']
    ctx.obj['config'] = cfg_gen(envname) if envname is not None else cfg_gen()
    ctx.obj['config'].seed = seed
    if len(override) > 0:
        import builtins
        try:
            exec(override, builtins.__dict__, {'config': ctx.obj['config']})
        except Exception as e:
            print('!!! Your override string \'{}\' contains an error !!!'.format(override))
            raise e


@rainy_cli.command(help='Train agents')
@click.pass_context
@click.option('--comment', type=str, default=None, help='Comment wrote to fingerprint.txt')
@click.option('--prefix', type=str, default='', help='Prefix of the log directory')
def train(ctx: dict, comment: Optional[str], prefix: str) -> None:
    c = ctx.obj['config']
    scr = ctx.obj['script_path']
    if scr:
        c.logger.set_dir_from_script_path(scr, comment=comment, prefix=prefix)
    c.logger.set_stderr()
    ag = ctx.obj['make_agent'](c)
    run.train_agent(ag)
    if mpi.IS_MPI_ROOT:
        print("random play: {}, trained: {}".format(ag.random_episode(), ag.eval_episode()))


@rainy_cli.command(help='Run the random agent and show its result')
@click.option('--save', is_flag=True, help='Save actions')
@click.option('--render', is_flag=True, help='Render the agent')
@click.option('--replay', is_flag=True, help='Show replay(works with special environments)')
@click.option('--action-file', type=str,
              default='best-actions.json', help='Name of the action file')
@click.pass_context
def random(ctx: dict, save: bool, render: bool, replay: bool, action_file: str) -> None:
    c = ctx.obj['config']
    ag = ctx.obj['make_agent'](c)
    action_file = fname if save else None
    run.random_agent(ag, render=render, replay=replay, action_file=action_file)


@rainy_cli.command(help='Given the Load a save file and restart training')
@click.pass_context
@click.argument('logdir')
@click.option('--model', type=str, default=run.SAVE_FILE_DEFAULT, help='Name of the save file')
@click.option('--additional-steps', type=int,
              default=100, help='The number of  additional training steps')
def retrain(ctx: dict, logdir: str, model: str, additional_steps: int) -> None:
    c = ctx.obj['config']
    log = c.logger.retrive(logdir)
    c.logger.set_stderr()
    ag = ctx.obj['make_agent'](c)
    run.retrain_agent(ag, log, load_file_name=model, additional_steps=additional_steps)
    print("random play: {}, trained: {}".format(ag.random_episode(), ag.eval_episode()))


@rainy_cli.command(help='Load a save file and evaluate the agent')
@click.argument('logdir')
@click.option('--model', type=str, default=run.SAVE_FILE_DEFAULT, help='Name of the save file')
@click.option('--render', is_flag=True, help='Render the agent')
@click.option('--replay', is_flag=True, help='Show replay(works with special environments)')
@click.option('--action-file', type=str,
              default='best-actions.json', help='Name of the action file')
@click.pass_context
def eval(ctx: dict, logdir: str, model: str, render: bool, replay: bool, action_file: str) -> None:
    c = ctx.obj['config']
    ag = ctx.obj['make_agent'](c)
    run.eval_agent(
        ag,
        logdir,
        load_file_name=model,
        render=render,
        replay=replay,
        action_file=action_file
    )


@rainy_cli.command(help='Open an ipython shell with rainy imported')
@click.option('--logdir', type=str, help='Name of the directly where the log file')
@click.option('--vi-mode', is_flag=True, help='Open ipython shell with vi-mode enabled')
@click.pass_context
def ipython(ctx: dict, logdir: Optional[str], vi_mode: bool) -> None:
    config, make_agent = ctx.obj['config'], ctx.obj['make_agent']  # noqa
    if logdir is not None:
        log = ExperimentLog(logdir)  # noqa
    else:
        open_log = ExperimentLog  # noqa
    try:
        from ptpython.ipython import embed
        del ctx, logdir
        import rainy  # noqa
        embed(vi_mode=vi_mode)
    except ImportError:
        print("To use ipython mode, install ipython and ptpython first.")


def run_cli(
        config_gen: Callable[..., Config],
        agent_gen: Callable[[Config], Agent],
        script_path: Optional[str] = None
) -> rainy_cli:
    obj = {
        'config_gen': config_gen,
        'make_agent': agent_gen,
        'script_path': script_path
    }
    rainy_cli(obj=obj)



from abc import ABC, abstractmethod
from torch import nn, Tensor
from torch.distributions import Bernoulli, Categorical, Distribution, Normal
from torch.nn import functional as F
from typing import Any, Optional
from ..prelude import Array, Self
from ..utils import Device


class Policy(ABC):
    def __init__(self, dist: Distribution) -> None:
        self.dist = dist
        self._action: Optional[Tensor] = None

    def action(self) -> Tensor:
        if self._action is None:
            self._action = self.dist.sample().detach()
        return self._action.squeeze()

    def action_as_np(self) -> Array:
        return self.action().cpu().numpy()

    def set_action(self, action: Tensor) -> None:
        self._action = action

    @abstractmethod
    def __getitem__(self, idx: Any) -> Self:
        pass

    @abstractmethod
    def best_action(self) -> Tensor:
        pass

    @abstractmethod
    def log_prob(self) -> Tensor:
        pass

    @abstractmethod
    def entropy(self) -> Tensor:
        pass


class BernoulliPolicy(Policy):
    def best_action(self) -> Tensor:
        return self.dist.probs > 0.5

    def log_prob(self) -> Tensor:
        return self.dist.log_prob(self.action())

    def entropy(self) -> Tensor:
        return self.dist.entropy()

    def __getitem__(self, idx: Any) -> Self:
        return BernoulliPolicy(Bernoulli(logits=self.dist.logits[idx]))


class CategoricalPolicy(Policy):
    def best_action(self) -> Tensor:
        return self.dist.probs.argmax(dim=-1)

    def log_prob(self) -> Tensor:
        return self.dist.log_prob(self.action())

    def entropy(self) -> Tensor:
        return self.dist.entropy()

    def __getitem__(self, idx: Any) -> Self:
        return CategoricalPolicy(Categorical(logits=self.dist.logits[idx]))


class GaussianPolicy(Policy):
    def best_action(self) -> Tensor:
        return self.dist.mean

    def entropy(self) -> Tensor:
        return self.dist.entropy().sum(-1)

    def log_prob(self) -> Tensor:
        return self.dist.log_prob(self.action()).sum(-1)

    def __getitem__(self, idx: Any) -> Self:
        return GaussianPolicy(Normal(self.dist.mean[idx], self.dist.stddev[idx]))


class PolicyDist(ABC, nn.Module):
    def __init__(self, action_dim: int, *args, **kwargs) -> None:
        super().__init__()
        self.action_dim = action_dim

    @property
    def input_dim(self) -> int:
        return self.action_dim

    @abstractmethod
    def forward(self, t: Tensor) -> Policy:
        pass


class BernoulliDist(PolicyDist):
    """Bernoulli policy with no learnable parameter
    """
    def forward(self, x: Tensor) -> Policy:
        return BernoulliPolicy(Bernoulli(logits=x))


class CategoricalDist(PolicyDist):
    """Categorical policy with no learnable parameter
    """
    def forward(self, x: Tensor) -> Policy:
        return CategoricalPolicy(Categorical(logits=x))


class GaussinanDist(PolicyDist):
    """Gaussian policy which takes both mean and stdev as inputs
    """
    @property
    def input_dim(self) -> int:
        return self.action_dim * 2

    def forward(self, x: Tensor) -> Policy:
        size = x.size(1) // 2
        mean, stddev = x[:, :size], x[:, size:]
        return GaussianPolicy(Normal(mean, F.softplus(stddev)))


class SeparateStdGaussianDist(PolicyDist):
    """Gaussian policy which takes only mean as an input, and has a standard deviation
       independent with states, as a lernable parameter.
    """
    def __init__(self, action_dim: int, device: Device) -> None:
        super().__init__(action_dim)
        self.stddev = nn.Parameter(device.zeros(action_dim))

    def forward(self, x: Tensor) -> Policy:
        return GaussianPolicy(Normal(x, F.softplus(self.stddev)))



"""Implementation of deque using 2 lists
"""
from collections.abc import Sequence
from typing import Any, List, Optional
from ..utils.sample import sample_indices


class ArrayDeque(Sequence):
    def __init__(self, capacity: Optional[int] = None, init_list: List[Any] = []) -> None:
        self.capacity = capacity
        self.front: List[Any] = []
        if capacity:
            init_list = init_list[-capacity:]
        self.back: List[Any] = init_list

    def __len__(self):
        return len(self.front) + len(self.back)

    def __getitem__(self, i: int) -> Any:  # type: ignore
        front_len = len(self.front)
        if i < front_len:
            return self.front[~i]
        else:
            return self.back[i - front_len]

    def __setitem__(self, i: int, item: Any) -> None:
        front_len = len(self.front)
        if i < front_len:
            self.front[~i] = item
        else:
            self.back[i - front_len] = item

    def push_back(self, item: Any) -> None:
        self.back.append(item)
        if self.capacity and len(self) > self.capacity:
            self.pop_front()

    def push_front(self, item: Any) -> None:
        self.front.append(item)
        if self.capacity and len(self) > self.capacity:
            self.pop_back()

    def pop_back(self) -> Any:
        if not self.back:
            n = len(self)
            if n >= 2:
                self._balance()
            elif n >= 1:
                return self.front.pop()
            else:
                raise IndexError('[ArrayDeque::pop_back] Empty')
        return self.back.pop()

    def pop_front(self) -> Any:
        self._balance()
        if not self.front:
            raise IndexError('[ArrayDeque::pop_front] Empty')
        return self.front.pop()

    def clear(self) -> None:
        self.front.clear()
        self.back.clear()

    def _balance(self) -> None:
        front_len = len(self.front)
        back_len = len(self.back)
        if 3 * front_len < back_len:
            front_len_new = max((front_len + back_len) // 2, 1)
            ext_len = front_len_new - front_len
            self.front[:0] = [x for x in reversed(self.back[:ext_len])]
            del self.back[:ext_len]
        elif 3 * back_len < front_len:
            front_len_new = max((front_len + back_len) // 2, 1)
            ext_len = front_len - front_len_new
            self.back[:0] = [x for x in reversed(self.front[:ext_len])]
            del self.front[:ext_len]

    def sample(self, k):
        front_len = len(self.front)
        n = front_len + len(self.back)
        if n < k:
            raise ValueError('[ArrayDeque::sample] n < k')
        return [self.front[i] if i < front_len else self.back[i - front_len]
                for i in sample_indices(n, k)]

    def __repr__(self):
        return "ArrayDeque({})".format(str(list(self)))



import numpy as np
from typing import Any, Iterable, Tuple
from .ext import EnvSpec
from .parallel import ParallelEnv
from ..prelude import Action, Array, State
from ..utils import RunningMeanStd


class ParallelEnvWrapper(ParallelEnv[Action, State]):
    def __init__(self, penv: ParallelEnv) -> None:
        self.penv = penv

    def close(self) -> None:
        self.penv.close()

    def reset(self) -> Array[State]:
        return self.penv.reset()

    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array[State], Array[float], Array[bool], Array[Any]]:
        return self.penv.step(actions)

    def seed(self, seeds: Iterable[int]) -> None:
        self.penv.seed(seeds)

    @property
    def num_envs(self) -> int:
        return self.penv.num_envs

    @property
    def spec(self) -> EnvSpec:
        return self.penv.spec

    def extract(self, states: Iterable[State]) -> Array:
        return self.penv.extract(states)


class FrameStackParallel(ParallelEnvWrapper):
    """Parallel version of atari_wrappers.FrameStack
    """
    def __init__(self, penv: ParallelEnv, nstack: int = 4, dtype: type = np.float32) -> None:
        super().__init__(penv)
        idx = 0
        shape = self.penv.state_dim
        for dim in shape:
            if dim == 1:
                idx += 1
            else:
                break
        self.shape = (nstack, *self.penv.state_dim[idx:])
        self.obs = np.zeros((self.num_envs, *self.shape), dtype=dtype)

    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array, Array[float], Array[bool], Array[Any]]:
        state, reward, done, info = self.penv.step(actions)
        self.obs = np.roll(self.obs, shift=-1, axis=1)
        for i, _ in filter(lambda t: t[1], enumerate(done)):
            self.obs[i] = 0.0
        self.obs[:, -1] = self.extract(state).squeeze()
        return (self.obs, reward, done, info)

    def reset(self) -> Array[State]:
        self.obs.fill(0)
        state = self.penv.reset()
        self.obs[:, -1] = self.extract(state).squeeze()
        return self.obs

    @property
    def state_dim(self) -> Tuple[int, ...]:
        return self.shape


class NormalizeObs(ParallelEnvWrapper[Action, Array[float]]):
    def __init__(self, penv: ParallelEnv, obs_clip: float = 10.) -> None:
        super().__init__(penv)
        self.obs_clip = obs_clip
        self._rms = RunningMeanStd(shape=self.state_dim)
        self.training_mode = True

    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array[Array[float]], Array[float], Array[bool], Array[Any]]:
        state, reward, done, info = self.penv.step(actions)
        return self._filter_obs(state), reward, done, info

    def _filter_obs(self, obs: Array[Array[float]]) -> Array[Array[float]]:
        if self.training_mode:
            self._rms.update(obs)  # type: ignore
        obs = np.clip((obs - self._rms.mean) / self._rms.std(), -self.obs_clip, self.obs_clip)
        return obs

    def reset(self) -> Array[Array[float]]:
        obs = self.penv.reset()
        return self._filter_obs(obs)


class NormalizeReward(ParallelEnvWrapper[Action, State]):
    def __init__(self, penv: ParallelEnv, reward_clip: float = 10., gamma: float = 0.99) -> None:
        super().__init__(penv)
        self.reward_clip = reward_clip
        self.gamma = gamma
        self._rms = RunningMeanStd(shape=())
        self.ret = np.zeros(self.num_envs)

    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array[State], Array[float], Array[bool], Array[Any]]:
        state, reward, done, info = self.penv.step(actions)
        self.ret = self.ret * self.gamma + reward
        self._rms.update(self.ret)
        reward = np.clip(reward / self._rms.std(), -self.reward_clip, self.reward_clip)
        self.ret[done] = 0.0
        return state, reward, done, info

    def reset(self) -> Array[State]:
        self.ret = np.zeros(self.num_envs)
        return self.penv.reset()



from abc import ABC, abstractmethod
import torch
from torch import LongTensor, Tensor
from torch.optim import Optimizer
from ..net.value import QFunction
from ..prelude import Array


class Cooler(ABC):
    @abstractmethod
    def __call__(self) -> float:
        pass

    def lr_decay(self, optim: Optimizer) -> None:
        for param_group in optim.param_groups:
            if 'lr' in param_group:
                param_group['lr'] = self.__call__()


class LinearCooler(Cooler):
    """decrease epsilon linearly, from initial to minimal, via `max_step` steps
    """
    def __init__(self, initial: float, minimal: float, max_step: int) -> None:
        if initial < minimal:
            raise ValueError(
                f'[LinearCooler.__init__] the minimal value({minimal})'
                ' is bigger than the initial value {initial}'
            )
        self.base = initial - minimal
        self.minimal = minimal
        self.max_step = max_step
        self.left = max_step

    def __call__(self) -> float:
        self.left = max(0, self.left - 1)
        return (float(self.left) / self.max_step) * self.base + self.minimal


class DummyCooler(Cooler):
    """Do nothing
    """
    def __init__(self, initial: float, *args) -> None:
        self.initial = initial

    def __call__(self) -> float:
        return self.initial


class Explorer(ABC):
    def select_action(self, state: Array, qfunc: QFunction) -> LongTensor:
        return self.select_from_value(qfunc.action_values(state).detach())

    @abstractmethod
    def select_from_value(self, value: Tensor) -> LongTensor:
        pass


class Greedy(Explorer):
    """deterministic greedy policy
    """
    def select_from_value(self, value: Tensor) -> LongTensor:
        return value.argmax(-1)  # type: ignore


class EpsGreedy(Explorer):
    """ε-greedy policy
    """
    def __init__(self, epsilon: float, cooler: Cooler) -> None:
        self.epsilon = epsilon
        self.cooler = cooler

    def select_from_value(self, value: Tensor) -> LongTensor:
        old_eps = self.epsilon
        self.epsilon = self.cooler()
        out_shape, action_dim = value.shape[:-1], value.size(-1)
        greedy = value.argmax(-1).view(-1).cpu()
        random = torch.randint(action_dim, value.shape[:-1]).view(-1)
        res = torch.where(torch.zeros(out_shape).view(-1) < old_eps, random, greedy)
        return res.reshape(out_shape).to(value.device)  # type: ignore



from torch import nn, Tensor
from typing import Callable, Iterable, Tuple, Union
from ..utils.device import Device

NetFn = Callable[[Tuple[int, ...], int, Device], nn.Module]
Params = Union[Iterable[Tensor], dict]



import torch
from torch import LongTensor, nn, Tensor
from typing import List, Union
from numpy import ndarray
from ..lib import mpi
from ..prelude import Self


class Device:
    """Utilities for handling devices
    """
    def __init__(self, use_cpu: bool = False, gpu_indices: List[int] = []) -> None:
        """
        :param gpu_limits: list of gpus you allow PyTorch to use
        """
        super().__init__()
        if use_cpu or not torch.cuda.is_available():
            self.__use_cpu()
        else:
            self.gpu_indices = gpu_indices if gpu_indices else self.__all_gpu()
            self.device = torch.device('cuda:{}'.format(self.gpu_indices[0]))

    def split(self) -> Self:
        """If self has multiple GPUs, split them into two devices.
        """
        num_gpus = len(self.gpu_indices)
        if num_gpus < 2:
            return self
        else:
            index = num_gpus // 2
            gpu_indices = self.gpu_indices[index:]
            self.gpu_indices = self.gpu_indices[:index]
            return Device(gpu_indices=gpu_indices)

    @property
    def unwrapped(self) -> torch.device:
        return self.device

    def tensor(self, arr: Union[ndarray, List[ndarray], Tensor], dtype=torch.float32) -> Tensor:
        """Convert numpy array or PyTorch Tensor into Tensor on main_device
        :param x: ndarray or Tensor you want to convert
        :return: Tensor
        """
        t = type(arr)
        if t is Tensor:
            return arr.to(device=self.device)  # type: ignore
        elif t is ndarray or t is list:
            return torch.tensor(arr, device=self.device, dtype=dtype)
        else:
            raise ValueError('arr must be ndarray or list or tensor')

    def zeros(self, size: Union[int, tuple], dtype: torch.dtype = torch.float32) -> Tensor:
        return torch.zeros(size, device=self.device, dtype=dtype)

    def ones(self, size: Union[int, tuple], dtype: torch.dtype = torch.float32) -> Tensor:
        return torch.ones(size, device=self.device, dtype=dtype)

    def data_parallel(self, module: nn.Module) -> nn.DataParallel:
        return nn.DataParallel(module, device_ids=self.gpu_indices)

    def is_multi_gpu(self) -> bool:
        return len(self.gpu_indices) > 1

    def indices(self, end: int, start: int = 0) -> LongTensor:
        res = torch.arange(start=start, end=end, device=self.device, dtype=torch.long)
        return res  # type: ignore

    def __all_gpu(self) -> List[int]:
        ngpus = torch.cuda.device_count()
        size, rank = mpi.local_size_and_rank()
        if size > 1:
            if ngpus < size:
                raise ValueError(f'Too many local processes {size} for {ngpus} gpus')
            return [rank]
        return list(range(ngpus))

    def __use_cpu(self) -> None:
        self.gpu_indices = []
        self.device = torch.device('cpu')

    def __repr__(self) -> str:
        return str(self.device)



import torch
from torch import Tensor
from typing import Generic, NamedTuple, Iterator, List, Optional
from ..envs import ParallelEnv, State
from ..net import DummyRnn, Policy, RnnBlock, RnnState
from ..utils import Device
from ..utils.misc import normalize_
from ..utils.sample import FeedForwardBatchSampler, RecurrentBatchSampler
from ..prelude import Array, Index


class RolloutStorage(Generic[State]):
    def __init__(self, nsteps: int, nworkers: int, device: Device) -> None:
        self.states: List[Array[State]] = []
        self.rewards: List[Array[float]] = []
        self.masks: List[Tensor] = [device.zeros(nworkers)]
        self.rnn_states: List[RnnState] = []
        self.policies: List[Policy] = []
        self.values: List[Tensor] = []
        self.returns: Tensor = device.zeros((nsteps + 1, nworkers))
        self.advs: Tensor = device.zeros((nsteps + 1, nworkers))
        self.batch_values: Tensor = device.zeros((nsteps, nworkers))
        self.nsteps = nsteps
        self.nworkers = nworkers
        self.device = device
        self._gae_used = False

    def initialized(self) -> bool:
        return len(self.states) != 0

    def set_initial_state(
            self,
            state: Array[State],
            rnn_state: RnnState = DummyRnn.DUMMY_STATE
    ) -> None:
        self.states.append(state)
        self.rnn_states.append(rnn_state)

    def push(
            self,
            state: Array[State],
            reward: Array[float],
            mask: Array[bool],
            rnn_state: RnnState = DummyRnn.DUMMY_STATE,
            policy: Optional[Policy] = None,
            value: Optional[Tensor] = None,
    ) -> None:
        assert self.states, '[RolloutStorage.push] Call set_initial_state first'
        self.states.append(state)
        self.rewards.append(reward)
        self.masks.append(self.device.tensor(1.0 - mask))
        self.rnn_states.append(rnn_state)
        if policy is not None:
            self.policies.append(policy)
        if value is not None:
            self.values.append(value.to(self.device.unwrapped))

    def reset(self) -> None:
        self.masks = [self.masks[-1]]
        self.states = [self.states[-1]]
        self.rnn_states = [self.rnn_states[-1]]
        self.rewards = []
        self.policies = []
        self.values = []

    def batch_states(self, penv: ParallelEnv) -> Tensor:
        states = [self.device.tensor(penv.extract(s)) for s in self.states[:-1]]
        return torch.cat(states, dim=0)

    def batch_actions(self) -> Tensor:
        return torch.cat([p.action() for p in self.policies])

    def batch_masks(self) -> Tensor:
        return torch.cat(self.masks[:-1])

    def batch_log_probs(self) -> Tensor:
        return torch.cat([p.log_prob() for p in self.policies])

    def _calc_ret_common(self, next_value: Tensor) -> Tensor:
        self.returns[-1] = next_value
        self.values.append(next_value)
        torch.stack(self.values[:self.nsteps], dim=0, out=self.batch_values)
        return self.device.tensor(self.rewards)

    def calc_ac_returns(self, next_value: Tensor, gamma: float) -> None:
        rewards = self._calc_ret_common(next_value)
        for i in reversed(range(self.nsteps)):
            self.returns[i] = self.returns[i + 1] * gamma * self.masks[i + 1] + rewards[i]
        self.advs[:-1] = self.returns[:-1] - self.batch_values

    def calc_gae_returns(self, next_value: Tensor, gamma: float, lambda_: float) -> None:
        self._gae_used = True
        rewards = self._calc_ret_common(next_value)
        self.advs.fill_(0.0)
        for i in reversed(range(self.nsteps)):
            td_error = \
                rewards[i] + gamma * self.values[i + 1] * self.masks[i + 1] - self.values[i]
            self.advs[i] = td_error + gamma * lambda_ * self.masks[i] * self.advs[i + 1]
            self.returns[i] = self.advs[i] + self.values[i]


class RolloutBatch(NamedTuple):
    states: Tensor
    actions: Tensor
    masks: Tensor
    returns: Tensor
    values: Tensor
    old_log_probs: Tensor
    advantages: Tensor
    rnn_init: RnnState


class RolloutSampler:
    def __init__(
            self,
            storage: RolloutStorage,
            penv: ParallelEnv,
            minibatch_size: int,
            rnn: RnnBlock = DummyRnn(),
            adv_normalize_eps: Optional[float] = None,
    ) -> None:
        """Create a batch sampler from storage for feed forward network.
        adv_normalize_eps is adpoted from the PPO implementation in OpenAI baselines.
        It looks for reducing the variance of advantages, but not sure.
        """
        self.nworkers = storage.nworkers
        self.nsteps = storage.nsteps
        if minibatch_size >= self.nworkers * self.nsteps:
            raise ValueError(
                'PPO requires minibatch_size <= nsteps * nworkers, but '
                'minibatch_size: {}, nsteps: {}, nworkers: {} was passed.'
                .format(minibatch_size, storage.nsteps, storage.nworkers)
            )
        self.minibatch_size = minibatch_size
        self.states = storage.batch_states(penv)
        self.actions = storage.batch_actions()
        self.masks = storage.batch_masks()
        self.returns = storage.returns[:-1].flatten()
        self.values = storage.batch_values.flatten()
        self.old_log_probs = storage.batch_log_probs()
        self.rnn_init = storage.rnn_states[0]
        self.advantages = storage.advs[:-1].flatten()
        if adv_normalize_eps is not None:
            normalize_(self.advantages, adv_normalize_eps)

    def _make_batch(self, i: Index) -> RolloutBatch:
        return RolloutBatch(
            self.states[i],
            self.actions[i],
            self.masks[i],
            self.returns[i],
            self.values[i],
            self.old_log_probs[i],
            self.advantages[i],
            self.rnn_init[i[:len(i) // self.nsteps]]
        )

    def __iter__(self) -> Iterator[RolloutBatch]:
        sampler_cls = FeedForwardBatchSampler \
            if self.rnn_init is DummyRnn.DUMMY_STATE else RecurrentBatchSampler
        for i in sampler_cls(self.nsteps, self.nworkers, self.minibatch_size):  # type: ignore
            yield self._make_batch(i)




from .array_deque import ArrayDeque
from collections import deque
import random


def test_deque_push_back() -> None:
    deq = ArrayDeque(capacity=10)
    for i in range(14):
        deq.push_back(i)

    for i in range(4, 14):
        assert deq[i - 4] == i


def test_deque_push_front() -> None:
    deq = ArrayDeque(capacity=10)
    for i in range(14):
        deq.push_front(i)
    print(deq)
    for i in range(4, 14):
        assert deq[13 - i] == i


def test_deque_stress() -> None:
    mydeq = ArrayDeque()
    deq = deque()
    N = 10000
    for i in range(N):
        num = random.randint(10, 1000000000)
        deq.append(num)
        mydeq.push_back(num)

    for i in range(N // 2):
        cond = random.randint(1, 4)
        num = random.randint(10, 1000000000)
        if cond is 1:
            deq.append(num)
            mydeq.push_back(num)
        elif cond is 2:
            deq.appendleft(num)
            mydeq.push_front(num)
        elif cond is 3:
            assert deq.pop() == mydeq.pop_back()
        else:
            assert deq.popleft() == mydeq.pop_front()


def test_deque_sample() -> None:
    deq = ArrayDeque()
    N = 10000
    for i in range(N):
        deq.push_back(i)
    K = 8000
    samples = deq.sample(K)
    samples = list(set(samples))
    assert len(samples) == K


def test_deque_sequence() -> None:
    deq = ArrayDeque(capacity=10)
    for i in range(14):
        deq.push_back(i)
    assert 12 in deq
    assert 18 not in deq
    assert min(deq) == 4
    assert len(deq) == 10
    assert deq.count(4) == 1


def test_deque_init_list() -> None:
    deq = ArrayDeque(capacity=10, init_list=[1, 2, 3])
    assert list(deq) == [1, 2, 3]
    deq = ArrayDeque(capacity=5, init_list=[1, 2, 3, 4, 5, 6, 7])
    assert list(deq) == [3, 4, 5, 6, 7]





import gym
from gym.wrappers import TimeLimit
from gym.spaces import Box
import numpy as np
from typing import Tuple, Union
from .atari_wrappers import LazyFrames
from ..prelude import Array


class TransposeObs(gym.ObservationWrapper):
    """Transpose & Scale image
    """
    def __init__(
            self,
            env: gym.Env,
            transpose: Tuple[int, int, int] = (2, 0, 1),
            scale: float = 255.0
    ) -> None:
        super().__init__(env)
        obs_shape = self.observation_space.shape
        self.observation_space: gym.Box = Box(
            low=self.observation_space.low[0, 0, 0],
            high=self.observation_space.high[0, 0, 0] / scale,
            shape=[obs_shape[i] for i in transpose],
            dtype=self.observation_space.dtype
        )
        self.scale = scale
        self.transpose = transpose

    def observation(self, observation: Union[Array, LazyFrames]):
        if isinstance(observation, LazyFrames):
            img = np.concatenate(observation._frames, axis=2).transpose(2, 0, 1)
        else:
            img = observation.transpose(*self.transpose)
        return img / self.scale


class AddTimeStep(gym.ObservationWrapper):
    def __init__(self, env: TimeLimit) -> None:
        super().__init__(env)
        obs = self.observation_space
        self.observation_space: gym.Box = Box(
            low=obs.low[0],
            high=obs.high[0],
            shape=[obs.shape[0] + 1],
            dtype=obs.dtype
        )

    def observation(self, obs: Array[float]) -> Array[float]:
        return np.append(obs, self.env._elapsed_steps)



"""Defines some reusable NN layers, named as 'Block'
"""
from abc import ABC, abstractmethod
import numpy as np
from torch import nn, Tensor
from typing import List, Sequence, Tuple
from .init import Initializer


class NetworkBlock(nn.Module, ABC):
    """Defines a NN block which returns 1-dimension Tensor
    """
    @property
    @abstractmethod
    def input_dim(self) -> Tuple[int, ...]:
        pass

    @property
    @abstractmethod
    def output_dim(self) -> int:
        pass


class DummyBlock(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return x


class LinearHead(NetworkBlock):
    """One FC layer
    """
    def __init__(self, input_dim: int, output_dim: int, init: Initializer = Initializer()) -> None:
        super().__init__()
        self.fc = init(nn.Linear(input_dim, output_dim))

    @property
    def input_dim(self) -> Tuple[int, ...]:
        return (self.fc.in_features, )

    @property
    def output_dim(self) -> int:
        return self.fc.out_features

    def forward(self, x: Tensor) -> Tensor:
        return self.fc(x)


class ConvBody(NetworkBlock):
    """Multiple CNN layers + FC
    """
    def __init__(
            self,
            cnns: Sequence[nn.Conv2d],
            fc: nn.Linear,
            input_dim: Tuple[int, int, int],
            activator: nn.Module = nn.ReLU(inplace=True),
            init: Initializer = Initializer(nonlinearity='relu'),
    ) -> None:
        super().__init__()
        self.cnns = init.make_list(cnns)
        self.fc = init(fc)
        self._input_dim = input_dim
        self.activator = activator

    @property
    def input_dim(self) -> Tuple[int, ...]:
        return self._input_dim

    @property
    def output_dim(self) -> int:
        return self.fc.out_features

    def forward(self, x: Tensor) -> Tensor:
        for cnn in self.cnns:
            x = self.activator(cnn(x))
        x = x.view(x.size(0), -1)
        x = self.activator(self.fc(x))
        return x


class DqnConv(ConvBody):
    """Convolutuion Network used in https://www.nature.com/articles/nature14236,
       but is parameterized for other usages.
    """
    def __init__(
            self,
            input_dim: Tuple[int, int, int],
            kernel_and_strides: Sequence[Tuple[int, int]] = [(8, 4), (4, 2), (3, 1)],
            hidden_channels: Sequence[int] = (32, 64, 64),
            output_dim: int = 512,
            activator: nn.Module = nn.ReLU(inplace=True),
            init: Initializer = Initializer(nonlinearity='relu'),
    ) -> None:
        in_channel, width, height = input_dim
        cnns, hidden = make_cnns(input_dim, kernel_and_strides, hidden_channels)
        fc = nn.Linear(hidden, output_dim)
        self._output_dim = output_dim
        super().__init__(cnns, fc, input_dim, activator=activator, init=init)


class ResBlock(nn.Sequential):
    def __init__(
            self,
            channel: int,
            stride: int = 1,
            use_batch_norm: bool = True,
    ) -> None:
        super().__init__(
            nn.ReLU(inplace=True),
            self._conv3x3(channel, channel, stride),
            self._batch_norm(use_batch_norm, channel),
            nn.ReLU(inplace=True),
            self._conv3x3(channel, channel, stride),
            self._batch_norm(use_batch_norm, channel),
        )

    @staticmethod
    def _batch_norm(use_batch_norm: bool, channel: int) -> nn.Module:
        if use_batch_norm:
            return nn.BatchNorm2d(channel)
        else:
            return DummyBlock()

    @staticmethod
    def _conv3x3(in_channel: int, out_channel: int, stride: int = 1) -> nn.Conv2d:
        return nn.Conv2d(
            in_channel,
            out_channel,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )

    def forward(self, x: Tensor) -> Tensor:
        residual = x
        out = super().forward(x)
        return out + residual


class ResNetBody(NetworkBlock):
    """Convolutuion Network used in IMPALA
    """
    def __init__(
            self,
            input_dim: Tuple[int, int, int],
            channels: List[int] = [16, 32, 32],
            maxpools: List[tuple] = [(3, 2, 1)] * 3,
            use_batch_norm: bool = True,
            fc_out: int = 256,
            init: Initializer = Initializer(nonlinearity='relu'),
    ) -> None:
        def layer(in_channel: int, out_channel: int, maxpool: tuple) -> nn.Sequential:
            return nn.Sequential(
                ResBlock._conv3x3(in_channel, out_channel),
                ResBlock._batch_norm(use_batch_norm, out_channel),
                nn.MaxPool2d(*maxpool),
                ResBlock(out_channel, use_batch_norm=use_batch_norm),
                ResBlock(out_channel, use_batch_norm=use_batch_norm)
            )

        super().__init__()
        self._input_dim = input_dim
        self.res_blocks = init.make_list([
            layer(*t) for t in zip([input_dim[0]] + channels, channels, maxpools)
        ])
        self.relu = nn.ReLU(inplace=True)
        conved = calc_cnn_hidden(maxpools, *input_dim[1:])
        fc_in = np.prod((channels[-1], *conved))
        self.fc = nn.Linear(fc_in, fc_out)

    def forward(self, x: Tensor) -> Tensor:
        for block in self.res_blocks:
            x = block(x)
        x = self.relu(x)
        x = self.fc(x.view(x.size(0), -1))
        return self.relu(x)

    @property
    def input_dim(self) -> Tuple[int, ...]:
        return self._input_dim

    @property
    def output_dim(self) -> int:
        return self.fc.out_features


class FcBody(NetworkBlock):
    def __init__(
            self,
            input_dim: int,
            units: List[int] = [64, 64],
            activator: nn.Module = nn.ReLU(inplace=True),
            init: Initializer = Initializer()
    ) -> None:
        super().__init__()
        self._input_dim = input_dim
        dims = [input_dim] + units
        self.layers = init.make_list(map(lambda i, o: nn.Linear(i, o), *zip(dims[:-1], dims[1:])))
        self.activator = activator
        self.dims = dims

    def forward(self, x: Tensor) -> Tensor:
        for layer in self.layers:
            x = self.activator(layer(x))
        return x

    @property
    def input_dim(self) -> Tuple[int, ...]:
        return (self.dims[0],)

    @property
    def output_dim(self) -> int:
        return self.dims[-1]


def make_cnns(
        input_dim: Tuple[int, int, int],
        params: Sequence[tuple],
        hidden_channels: Sequence[int]
) -> Tuple[List[nn.Conv2d], int]:
    """Make a list of CNNs from lists of parameters.
    """
    channel, width, height = input_dim
    hiddens = list(hidden_channels)
    res = []
    for ic, oc, param in zip([channel] + hiddens, hiddens, params):
        res.append(nn.Conv2d(ic, oc, *param))
    hidden = (hidden_channels[-1], *calc_cnn_hidden(params, width, height))
    return res, np.prod(hidden)


def calc_cnn_hidden(params: Sequence[tuple], width: int, height: int) -> Tuple[int, int]:
    """Calcurate hidden dim of a CNN.
       See https://pytorch.org/docs/stable/nn.html#torch.nn.Conv2d for detail.
    """
    for param in params:
        kernel, stride, padding = param if len(param) > 2 else (*param, 0)
        width = (width - kernel + 2 * padding) // stride + 1
        height = (height - kernel + 2 * padding) // stride + 1
    assert width > 0 and height > 0, 'Convolution makes dim < 0!!!'
    return width, height



import torch
from torch import Tensor
from typing import Tuple
from .a2c import A2cAgent
from ..lib.rollout import RolloutSampler
from ..lib import mpi
from ..config import Config
from ..envs import State
from ..net import ActorCriticNet, Policy
from ..prelude import Array


class PpoAgent(A2cAgent):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.net: ActorCriticNet = config.net('actor-critic')  # type: ignore
        self.optimizer = config.optimizer(self.net.parameters())
        self.lr_cooler = config.lr_cooler(self.optimizer.param_groups[0]['lr'])
        self.clip_cooler = config.clip_cooler()
        self.clip_eps = config.ppo_clip
        nbatchs = (self.config.nsteps * self.config.nworkers) // self.config.ppo_minibatch_size
        self.num_updates = self.config.ppo_epochs * nbatchs
        mpi.setup_models(self.net)
        self.optimizer = mpi.setup_optimizer(self.optimizer)

    def members_to_save(self) -> Tuple[str, ...]:
        return 'net', 'clip_eps', 'clip_cooler', 'optimizer'

    def _policy_loss(self, policy: Policy, advantages: Tensor, old_log_probs: Tensor) -> Tensor:
        prob_ratio = torch.exp(policy.log_prob() - old_log_probs)
        surr1 = prob_ratio * advantages
        surr2 = prob_ratio.clamp(1.0 - self.clip_eps, 1.0 + self.clip_eps) * advantages
        return -torch.min(surr1, surr2).mean()

    def _value_loss(self, value: Tensor, old_value: Tensor, returns: Tensor) -> Tensor:
        """Clip value function loss.
        OpenAI baselines says it reduces variability during Critic training... but I'm not sure.
        """
        unclipped_loss = (value - returns).pow(2)
        if not self.config.ppo_value_clip:
            return unclipped_loss.mean()
        value_clipped = old_value + (value - old_value).clamp(-self.clip_eps, self.clip_eps)
        clipped_loss = (value_clipped - returns).pow(2)
        return torch.max(unclipped_loss, clipped_loss).mean()

    def nstep(self, states: Array[State]) -> Array[State]:
        for _ in range(self.config.nsteps):
            states = self._one_step(states)

        with torch.no_grad():
            next_value = self.net.value(*self._network_in(states))

        if self.config.use_gae:
            self.storage.calc_gae_returns(
                next_value,
                self.config.discount_factor,
                self.config.gae_lambda,
            )
        else:
            self.storage.calc_ac_returns(next_value, self.config.discount_factor)
        p, v, e = (0.0, 0.0, 0.0)
        sampler = RolloutSampler(
            self.storage,
            self.penv,
            self.config.ppo_minibatch_size,
            rnn=self.net.recurrent_body,
            adv_normalize_eps=self.config.adv_normalize_eps,
        )
        for _ in range(self.config.ppo_epochs):
            for batch in sampler:
                policy, value, _ = self.net(batch.states, batch.rnn_init, batch.masks)
                policy.set_action(batch.actions)
                policy_loss = self._policy_loss(policy, batch.advantages, batch.old_log_probs)
                value_loss = self._value_loss(value, batch.values, batch.returns)
                entropy_loss = policy.entropy().mean()
                self.optimizer.zero_grad()
                (policy_loss
                 + self.config.value_loss_weight * 0.5 * value_loss
                 - self.config.entropy_weight * entropy_loss).backward()
                mpi.clip_and_step(self.net, self.config.grad_clip, self.optimizer)
                p, v, e = p + policy_loss.item(), v + value_loss.item(), e + entropy_loss.item()

        self.lr_cooler.lr_decay(self.optimizer)
        self.clip_eps = self.clip_cooler()
        self.storage.reset()

        p, v, e = (x / self.num_updates for x in (p, v, e))
        self.report_loss(policy_loss=p, value_loss=v, entropy_loss=e)
        return states



from functools import partial
from typing import Iterable, Optional, Union
import torch
from torch import nn, Tensor

InitFn = partial


def uniform(mean: float = 0., var: float = 1.) -> InitFn:
    return partial(nn.init.uniform_, a=mean, b=var)


def orthogonal(gain: float = 1.) -> InitFn:
    return partial(nn.init.orthogonal_, gain=gain)


def kaiming_normal(nonlinearity: str = 'relu') -> InitFn:
    return partial(nn.init.kaiming_normal_, nonlinearity=nonlinearity)


def kaiming_uniform(nonlinearity: str = 'relu') -> InitFn:
    return partial(nn.init.kaiming_uniform_, nonlinearity=nonlinearity)


def constant(val: float) -> InitFn:
    return partial(nn.init.constant_, val=val)


def zero() -> InitFn:
    return partial(nn.init.constant_, val=0)


def lstm_bias(forget: float = 1.0, other: float = 0.0) -> InitFn:
    """Set forget bias and others separately.
    """
    @torch.no_grad()
    def __set_bias(t: Tensor) -> None:
        i = t.size(0) // 4
        t.fill_(other)
        t[i:2 * i].fill_(forget)
    return partial(__set_bias)


class Initializer:
    """Utility Class to initialize weight parameters of NN
    """
    def __init__(
            self,
            nonlinearity: Optional[str] = None,
            weight_init: InitFn = orthogonal(),
            bias_init: InitFn = zero(),
            scale: float = 1.,
    ) -> None:
        """If nonlinearity is specified, use orthogonal
           with calucurated gain by torch.init.calculate_gain.
        """
        self.weight_init = weight_init
        if nonlinearity is not None:
            if 'gain' in self.weight_init.keywords:
                self.weight_init.keywords['gain'] = nn.init.calculate_gain(nonlinearity)
            elif 'nonlinearity' in self.weight_init.keywords:
                self.weight_init.keywords['nonlinearity'] = nonlinearity
            else:
                raise ValueError('{} doesn\'t have gain', self.weight_init)
        self.bias_init = bias_init
        self.scale = scale

    def __call__(self, mod: nn.Module) -> nn.Module:
        return self.__init_dispatch(mod)

    def make_list(self, mods: Iterable[nn.Module]) -> nn.ModuleList:
        return nn.ModuleList([self.__init_dispatch(mod) for mod in mods])

    def __init_dispatch(self, mod: nn.Module) -> nn.Module:
        if isinstance(mod, nn.Sequential) or isinstance(mod, nn.ModuleList):
            for child in mod.children():
                self.__init_dispatch(child)
        else:
            self.__init_mod(mod)
        return mod

    def __init_mod(self, mod: nn.Module) -> nn.Module:
        if isinstance(mod, nn.BatchNorm2d):
            return self.__init_batch_norm(mod)
        for name, param in mod.named_parameters():
            if 'weight' in name:
                self.weight_init(param)
            elif 'bias' in name:
                self.bias_init(param)
        return mod

    def __init_batch_norm(self, mod: nn.BatchNorm2d) -> nn.Module:
        mod.weight.data.fill_(1.0)
        mod.bias.data.zero_()
        return mod



from abc import ABC, abstractmethod
import multiprocessing as mp
from multiprocessing.connection import Connection
import numpy as np
from numpy import ndarray
from typing import Any, Callable, Generic, Iterable, Tuple
from .ext import EnvExt, EnvSpec
from ..prelude import Action, Array, State

EnvGen = Callable[[], EnvExt]


class ParallelEnv(ABC, Generic[Action, State]):
    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def reset(self) -> Array[State]:
        pass

    @abstractmethod
    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array[State], Array[float], Array[bool], Array[Any]]:
        pass

    @abstractmethod
    def seed(self, seeds: Iterable[int]) -> None:
        pass

    @property
    @abstractmethod
    def num_envs(self) -> int:
        pass

    @property
    @abstractmethod
    def spec(self) -> EnvSpec:
        pass

    @property
    def action_dim(self) -> int:
        return self.spec.action_dim

    @property
    def state_dim(self) -> Tuple[int, ...]:
        return self.spec.state_dim

    @property
    def use_reward_monitor(self) -> bool:
        return self.spec.use_reward_monitor

    @abstractmethod
    def extract(self, states: Iterable[State]) -> Array:
        """
        Convert Sequence of states to ndarray.
        """
        pass


class MultiProcEnv(ParallelEnv):
    def __init__(self, env_gen: EnvGen, nworkers: int) -> None:
        assert nworkers >= 2
        envs_tmp = [env_gen() for _ in range(nworkers)]
        self.to_array = envs_tmp[0].extract
        self._spec = envs_tmp[0].spec
        self.envs = [_ProcHandler(e) for e in envs_tmp]

    def close(self) -> None:
        for env in self.envs:
            env.close()

    def reset(self) -> Array[State]:
        for env in self.envs:
            env.reset()
        return np.array([env.recv() for env in self.envs])

    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array[State], Array[float], Array[bool], Array[Any]]:
        for env, action in zip(self.envs, actions):
            env.step(action)
        res = [env.recv() for env in self.envs]
        return tuple(map(np.array, zip(*res)))  # type: ignore

    def seed(self, seeds: Iterable[int]) -> None:
        for env, seed in zip(self.envs, seeds):
            env.seed(seed)

    @property
    def num_envs(self) -> int:
        return len(self.envs)

    @property
    def spec(self) -> EnvSpec:
        return self._spec

    def extract(self, states: Iterable[State]) -> ndarray:
        return np.asarray([self.to_array(s) for s in states])


class _ProcHandler:
    def __init__(self, env: EnvExt) -> None:
        self.pipe, worker_pipe = mp.Pipe()
        self.worker = _ProcWorker(env, worker_pipe)
        self.worker.start()

    def close(self) -> None:
        self.pipe.send((_ProcWorker.CLOSE, None))

    def reset(self) -> None:
        self.pipe.send((_ProcWorker.RESET, None))

    def seed(self, seed: int) -> None:
        self.pipe.send((_ProcWorker.SEED, seed))

    def step(self, action: Action) -> None:
        self.pipe.send((_ProcWorker.STEP, action))

    def recv(self) -> Any:
        return self.pipe.recv()


class _ProcWorker(mp.Process):
    CLOSE = 0
    RESET = 1
    SEED = 2
    STEP = 3

    def __init__(self, env: EnvExt, pipe: Connection) -> None:
        super(_ProcWorker, self).__init__()
        self.env = env
        self.pipe = pipe

    def run(self):
        while True:
            op, arg = self.pipe.recv()
            if op == self.STEP:
                self.pipe.send(self.env.step_and_reset(arg))
            elif op == self.RESET:
                self.pipe.send(self.env.reset())
            elif op == self.SEED:
                self.env.seed(arg)
            elif op == self.CLOSE:
                self.env.close()
                self.pipe.close()
                break


class DummyParallelEnv(ParallelEnv):
    def __init__(self, env_gen: EnvGen, nworkers: int) -> None:
        self.envs = [env_gen() for _ in range(nworkers)]

    def close(self) -> None:
        for env in self.envs:
            env.close()

    def reset(self) -> Array[State]:
        return np.array([e.reset() for e in self.envs])

    def step(
            self,
            actions: Iterable[Action]
    ) -> Tuple[Array[State], Array[float], Array[bool], Array[Any]]:
        res = [e.step_and_reset(a) for (a, e) in zip(actions, self.envs)]
        return tuple(map(np.array, zip(*res)))  # type: ignore

    def seed(self, seeds: Iterable[int]) -> None:
        for env, seed in zip(self.envs, seeds):
            env.seed(seed)

    @property
    def num_envs(self) -> int:
        return len(self.envs)

    @property
    def spec(self) -> EnvSpec:
        return self.envs[0].spec

    def extract(self, states: Iterable[State]) -> ndarray:
        return np.asarray([e.extract(s) for (s, e) in zip(states, self.envs)])



import numpy as np
from .sample import RecurrentBatchSampler, sample_indices


def test_sample_large():
    arr = sample_indices(100, 90)
    assert np.unique(arr).__len__() == 90


def test_sample_small():
    arr = sample_indices(100, 30)
    assert np.unique(arr).__len__() == 30


def test_recurrent_batch_sampler():
    NSTEPS, NWORKERS = 12, 16
    r = RecurrentBatchSampler(NSTEPS, NWORKERS, 24)
    for batch in r:
        batch = batch.reshape(NSTEPS, -1)
        for i in range(NSTEPS - 1):
            before = batch[i]
            after = batch[i + 1]
            np.testing.assert_array_equal(before + NWORKERS, after)



from numpy import ndarray
import gym
from typing import Callable, Optional
from .atari_wrappers import LazyFrames, make_atari, wrap_deepmind
from .ext import EnvExt, EnvSpec
from .monitor import RewardMonitor
from .obs_wrappers import AddTimeStep, TransposeObs
from .parallel import DummyParallelEnv, EnvGen, MultiProcEnv, ParallelEnv
from .parallel_wrappers import FrameStackParallel, NormalizeObs, \
    NormalizeReward, ParallelEnvWrapper
from ..prelude import Self, State


class AtariConfig:
    STYLES = ['deepmind', 'baselines', 'dopamine']

    def __init__(self) -> None:
        self.timelimit = True
        self.override_timelimit: Optional[int] = None
        self.v4 = False
        self.sticky_actions = False
        self.noop_reset = True
        self.episodic_life = False
        self.fire_reset = False
        self.clip_reward = True
        self.flicker_frame = False
        self.frame_stack = True

    @staticmethod
    def from_style(style: str) -> Self:
        if style not in AtariConfig.STYLES:
            raise ValueError('You have to choose a style from {}'.format(AtariConfig.STYLES))
        cfg = AtariConfig()
        if style == 'deepmind':
            pass
        elif style == 'baselines':
            cfg.fire_reset = True
        elif style == 'dopamine':
            cfg.timelimit = False
            cfg.v4 = True
            cfg.noop_reset = False
            cfg.episodic_life = True
        return cfg


class Atari(EnvExt):
    def __init__(self, name: str, cfg: AtariConfig = AtariConfig(), **kwargs) -> None:
        cfg.__dict__.update(kwargs)
        env = make_atari(
            name,
            timelimit=cfg.timelimit,
            override_timelimit=cfg.override_timelimit,
            v4=cfg.v4,
            sticky_actions=cfg.sticky_actions,
            noop_reset=cfg.noop_reset,
        )
        env = RewardMonitor(env)
        env = wrap_deepmind(
            env,
            episodic_life=cfg.episodic_life,
            fire_reset=cfg.fire_reset,
            clip_reward=cfg.clip_reward,
            flicker_frame=cfg.flicker_frame,
            frame_stack=cfg.frame_stack
        )
        env = TransposeObs(env)
        super().__init__(env)
        self.spec.use_reward_monitor = True

    @staticmethod
    def extract(obs: State) -> ndarray:
        if type(obs) is LazyFrames:
            return obs.__array__()  # type: ignore
        else:
            return obs


def atari_parallel(frame_stack: bool = True) -> Callable[[EnvGen, int], ParallelEnv]:
    def __wrap(env_gen: EnvGen, num_workers: int) -> ParallelEnv:
        penv: ParallelEnv = MultiProcEnv(env_gen, num_workers)
        if frame_stack:
            penv = FrameStackParallel(penv)
        return penv
    return __wrap


class ClassicalControl(EnvExt):
    def __init__(self, name: str = 'CartPole-v0', max_steps: int = 200) -> None:
        self.name = name
        super().__init__(gym.make(name))
        self._env._max_episode_steps = max_steps


class PyBullet(EnvExt):
    def __init__(
            self,
            name: str = 'Hopper',
            add_timestep: bool = False,
            nosuffix: bool = False
    ) -> None:
        self.name = name
        try:
            import pybullet_envs  # noqa
        except ImportError:
            raise ImportError('pybullet is not installed')
        if not nosuffix:
            name += 'BulletEnv-v0'
        env = gym.make(name)
        if add_timestep:
            env = AddTimeStep(env)
        super().__init__(RewardMonitor(env))
        self.viewer = None
        self.spec.use_reward_monitor = True

    def render(self, mode: str = 'human') -> Optional[ndarray]:
        if mode == 'human':
            arr = self._env.render('rgb_array')
            if self.viewer is None:
                from gym.envs.classic_control.rendering import SimpleImageViewer
                self.viewer = SimpleImageViewer()
            self.viewer.imshow(arr)  # type: ignore
            return None
        else:
            return self._env.render(mode)


def pybullet_parallel(
        normalize_obs: bool = True,
        normalize_reward: bool = True,
        obs_clip: float = 10.0,
        reward_clip: float = 10.0,
        gamma: float = 0.99
) -> Callable[[EnvGen, int], ParallelEnv]:
    def __wrap(env_gen: EnvGen, num_workers: int) -> ParallelEnv:
        penv: ParallelEnv = MultiProcEnv(env_gen, num_workers)
        if normalize_obs:
            penv = NormalizeObs(penv, obs_clip=obs_clip)
        if normalize_reward:
            penv = NormalizeReward(penv, reward_clip=reward_clip, gamma=gamma)
        return penv
    return __wrap




from abc import ABC, abstractmethod
from torch import nn, Tensor
from typing import Callable, Tuple, Union
from .actor_critic import policy_init
from .block import DqnConv, FcBody, LinearHead, NetworkBlock
from .policy import BernoulliDist, BernoulliPolicy, CategoricalDist, Policy, PolicyDist
from .prelude import NetFn
from ..prelude import Array
from ..utils import Device


class OptionCriticNet(nn.Module, ABC):
    """Network for option critic
    """
    @abstractmethod
    def opt_q(self, states: Union[Array, Tensor]) -> Tensor:
        pass

    @abstractmethod
    def forward(self, states: Union[Array, Tensor]) -> Tuple[Policy, Tensor, BernoulliPolicy]:
        pass


class SharedBodyOCNet(OptionCriticNet):
    """An Option Critic Net with shared body and separate π/β/Value heads
    """
    def __init__(
            self,
            body: NetworkBlock,
            actor_head: NetworkBlock,
            optq_head: NetworkBlock,
            beta_head: NetworkBlock,
            policy_dist: PolicyDist,
            device: Device = Device(),
    ) -> None:
        super().__init__()
        self.body = body
        self.actor_head = actor_head
        self.optq_head = optq_head
        self.beta_head = beta_head
        self.policy_dist = policy_dist
        self.beta_dist = BernoulliDist(1)
        self.num_options = optq_head.output_dim
        self.action_dim = actor_head.output_dim // self.num_options
        self.device = device
        self.to(device.unwrapped)

    @property
    def state_dim(self) -> Tuple[int, ...]:
        return self.body.input_dim

    def opt_q(self, states: Union[Array, Tensor]) -> Tensor:
        feature = self.body(self.device.tensor(states))
        return self.optq_head(feature)

    def forward(self, states: Union[Array, Tensor]) -> Tuple[Policy, Tensor, BernoulliPolicy]:
        feature = self.body(self.device.tensor(states))
        policy = self.actor_head(feature).view(-1, self.num_options, self.action_dim)
        opt_q = self.optq_head(feature)
        beta = self.beta_dist(self.beta_head(feature))
        return self.policy_dist(policy), opt_q, beta


def conv_shared(
        num_options: int = 8,
        policy: Callable[[int, Device], PolicyDist] = CategoricalDist,
        hidden_channels: Tuple[int, int, int] = (32, 64, 32),
        output_dim: int = 256,
        **kwargs
) -> NetFn:
    def _net(state_dim: Tuple[int, int, int], action_dim: int, device: Device) -> SharedBodyOCNet:
        body = DqnConv(state_dim, hidden_channels=hidden_channels, output_dim=output_dim, **kwargs)
        ac_head = LinearHead(body.output_dim, action_dim * num_options, policy_init())
        optq_head = LinearHead(body.output_dim, num_options)
        beta_head = LinearHead(body.output_dim, num_options)
        dist = policy(action_dim, device)
        return SharedBodyOCNet(body, ac_head, optq_head, beta_head, dist, device)
    return _net  # type: ignore


def fc_shared(
        num_options: int = 8,
        policy: Callable[[int, Device], PolicyDist] = CategoricalDist,
        **kwargs
) -> NetFn:
    def _net(state_dim: Tuple[int, ...], action_dim: int, device: Device) -> SharedBodyOCNet:
        body = FcBody(state_dim[0], **kwargs)
        ac_head = LinearHead(body.output_dim, action_dim * num_options, policy_init())
        optq_head = LinearHead(body.output_dim, num_options)
        beta_head = LinearHead(body.output_dim, num_options)
        dist = policy(action_dim, device)
        return SharedBodyOCNet(body, ac_head, optq_head, beta_head, dist, device)
    return _net  # type: ignore



from abc import ABC, abstractmethod
from contextlib import contextmanager
from enum import Enum
import math
import torch
from torch import nn, Tensor
from torch.nn import functional as F
from torch.optim import Optimizer, SGD
from typing import Any, Callable, Generator, List, Optional, Tuple, Union
import warnings
from ..net.prelude import Params


class Layer(Enum):
    LINEAR = 1
    CONV2D = 2


def get_layer(mod: nn.Module) -> Union[Layer, str]:
    name = type(mod).__name__
    if name == 'Linear':
        return Layer.LINEAR
    elif name == 'Conv2d':
        return Layer.CONV2D
    else:
        return name


def default_sgd(eta_max: float = 0.25, momentum: float = 0.9) -> Callable[[Params], Optimizer]:
    """Returns the SGD optimizer which has the default setting for used with K-FAC.
    """
    def _sgd(params: Params) -> Optimizer:
        return SGD(params, lr=eta_max * (1.0 - momentum), momentum=momentum)
    return _sgd


class NormScaler(ABC):
    @abstractmethod
    def scale(self, fisher_norm: float) -> float:
        pass

    def __call__(self, param_groups: List[dict], fisher_norm: float) -> None:
        scale = self.scale(fisher_norm)
        for group in param_groups:
            for param in group['params']:
                if param is not None:
                    param.grad.data.mul_(scale)


class SquaredFisherScaler(NormScaler):
    """Section 5 in https://jimmylba.github.io/papers/nsync.pdf
    """
    def __init__(self, eta_max: float = 0.25, delta: float = 0.001) -> None:
        self.eta_max2 = eta_max ** 2
        self.delta = delta

    def scale(self, fisher_norm: float) -> float:
        return min(1.0, math.sqrt(self.delta / (fisher_norm * self.eta_max2)))


class DiagonalScaler(NormScaler):
    """https://arxiv.org/abs/1705.09319
    """
    def __init__(self, mu: float = 0.001) -> None:
        self.mu = mu

    def scale(self, fisher_norm: float) -> float:
        return math.sqrt(1.0 / (fisher_norm + self.mu))


class DummyScaler(NormScaler):
    def scale(self, fisher_norm: float) -> float:
        return 1.0


class PreConditioner(ABC, Optimizer):
    pass


class KfacPreConditioner(PreConditioner):
    def __init__(
            self,
            net: nn.Module,
            damping: float = 1.0e-4,
            weight_decay: float = 0.0,
            tau: float = 100.0,
            eps: float = 1.0e-6,
            update_freq: int = 2,
            norm_scaler: NormScaler = SquaredFisherScaler(),
    ) -> None:
        self.gamma = (damping + weight_decay) ** 0.5
        self.beta = math.exp(-1.0 / tau)
        self.eps = eps
        self.update_freq = update_freq
        self.norm_scaler = norm_scaler
        self.params: List[dict] = []
        self._counter = 0
        self._save_grad = False
        for mod in net.modules():
            layer_type = get_layer(mod)
            if not isinstance(layer_type, Layer):
                warnings.warn('KfacPreConditioner doesn\'t support {}. Skipped it.'
                              .format(layer_type))
                continue
            mod.register_forward_pre_hook(self._save_x)
            mod.register_backward_hook(self._save_g)
            params = [mod.weight]
            if mod.bias is not None:
                params.append(mod.bias)
            self.params.append({'params': params, 'mod': mod, 'layer_type': layer_type})
        super().__init__(self.params, {})

    def with_saving_grad(self, f: Callable[[], Any]) -> Any:
        self._save_grad = True
        res = f()
        self._save_grad = False
        return res

    def _save_x(self, mod: nn.Module, input_: Tuple[Tensor, ...]) -> None:
        """Save the inputs of the layer
        """
        if mod.training:
            self.state[mod]['x'], *_ = input_

    def _save_g(self, mod: nn.Module, _grad_in: tuple, grad_out: Tuple[Tensor, ...]) -> None:
        """Save the output gradients of the layer
        """
        if mod.training and self._save_grad:
            grad, *_ = grad_out
            self.state[mod]['g'] = grad * grad.size(0)

    def step(self) -> None:
        fisher_norm = 0.0
        for group in self.param_groups:
            weight, bias = group['params']
            state = self.state[weight]
            self.__xxt(group, state)
            self.__ggt(group, state)
            if self._counter % self.update_freq == 0:
                self.__eigend(state)
            fisher_norm += self._update_params(weight, bias, group['layer_type'], state)
            del self.state[group['mod']]['x'], self.state[group['mod']]['g']
        self.norm_scaler(self.param_groups, fisher_norm)
        self._counter += 1

    def _update_params(
            self,
            weight: Tensor,
            bias: Optional[Tensor],
            layer: Layer,
            state: dict
    ) -> float:
        """Updates gradients
        """
        gw, gb = self.__fisher_grad(weight, bias, layer, state)
        fisher_norm = weight.grad.mul(gw).sum().item()
        weight.grad.data.copy_(gw)
        if bias is not None:
            fisher_norm += bias.grad.mul(gb).sum().item()
            bias.grad.data.copy_(gb)
        return fisher_norm

    def __fisher_grad(
            self,
            weight: Tensor,
            bias: Optional[Tensor],
            layer: Layer,
            state: dict
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Computes F^{-1}∇h
        """
        grad = weight.grad.data
        if layer is Layer.CONV2D:
            grad = grad.view(grad.size(0), -1)
        if bias is not None:
            grad = torch.cat([grad, bias.grad.data.view(-1, 1)], dim=1)
        v1 = torch.chain_matmul(state['vg'].t(), grad, state['vx'])
        v2 = v1.div_(state['eg*ex'].add(self.gamma))
        grad = torch.chain_matmul(state['vg'], v2, state['vx'].t())
        gb = None
        if bias is not None:
            gb = grad[:, -1].reshape(bias.shape)
            gw = grad[:, :-1]
        return gw.reshape(weight.shape), gb

    def __eigend(self, state: dict) -> None:
        """Computes eigen decomposition of E[xx] and E[gg]
        """
        ex, state['vx'] = torch.symeig(state['xxt'], eigenvectors=True)
        eg, state['vg'] = torch.symeig(state['ggt'], eigenvectors=True)
        state['eg*ex'] = torch.ger(eg.clamp_(min=self.eps), ex.clamp_(min=self.eps))

    def __xxt(self, group: dict, state: dict) -> None:
        """Computes E[xi xj]^T and memorize it
        """
        mod = group['mod']
        x = self.state[mod]['x']
        if group['layer_type'] is Layer.CONV2D:
            x = F.unfold(x, mod.kernel_size, padding=mod.padding, stride=mod.stride)
            x = x.data.transpose(1, 0).reshape(x.size(1), -1)
        else:
            x = x.data.t()
        if mod.bias is not None:
            x = torch.cat([x, torch.ones_like(x[:1])])
        self.__average(state, 'xxt', x, float(x.size(1)))

    def __ggt(self, group: dict, state: dict) -> None:
        """Computes E[gi gj]^T and memorize it
        """
        g = self.state[group['mod']]['g']
        if group['layer_type'] is Layer.CONV2D:
            g = g.data.transpose(1, 0).reshape(g.size(1), -1)
        else:
            g = g.data.t()
        self.__average(state, 'ggt', g, float(g.size(1)))

    def __average(self, state: dict, param: str, mat: Tensor, scale: float) -> None:
        """Computes the moving average X <- βX + (1-β)X'
        """
        if self._counter == 0:
            state[param] = torch.mm(mat, mat.t()).div_(scale)
        else:
            state[param].addmm_(
                mat1=mat,
                mat2=mat.t(),
                beta=self.beta,
                alpha=(1 - self.beta) / scale,
            )

    @contextmanager
    def save_grad(self) -> Generator[None, None, None]:
        self._save_grad = True
        yield
        self._save_grad = False



from abc import ABC, abstractmethod
import numpy as np
from torch import nn, Tensor
from typing import Callable, List, Optional, Tuple, Union
from .block import DqnConv, FcBody, ResNetBody, LinearHead, NetworkBlock
from .init import Initializer, orthogonal
from .policy import CategoricalDist, Policy, PolicyDist
from .prelude import NetFn
from .recurrent import DummyRnn, RnnBlock, RnnState
from ..prelude import Array
from ..utils import Device


class ActorCriticNet(nn.Module, ABC):
    """Network with Policy + Value Head
    """
    @property
    @abstractmethod
    def recurrent_body(self) -> RnnBlock:
        pass

    @property
    def is_recurrent(self) -> bool:
        return not isinstance(self.recurrent_body, DummyRnn)

    @abstractmethod
    def policy(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tuple[Policy, RnnState]:
        pass

    @abstractmethod
    def value(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tensor:
        pass

    @abstractmethod
    def forward(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tuple[Policy, Tensor, RnnState]:
        pass


class SharedBodyACNet(ActorCriticNet):
    """An Actor Critic network with common body + separate value/policy heads.
    Basically it's same as the one used in the Atari experimtent in the A3C paper.
    """
    def __init__(
            self,
            body: NetworkBlock,
            actor_head: NetworkBlock,  # policy
            critic_head: NetworkBlock,  # value
            policy_dist: PolicyDist,
            recurrent_body: RnnBlock = DummyRnn(),
            device: Device = Device(),
    ) -> None:
        assert body.output_dim == np.prod(actor_head.input_dim), \
            'body output and action_head input must have a same dimention'
        assert body.output_dim == np.prod(critic_head.input_dim), \
            'body output and action_head input must have a same dimention'
        super().__init__()
        self.body = body
        self.device = device
        self.body = body
        self.actor_head = actor_head
        self.critic_head = critic_head
        self.policy_dist = policy_dist
        self._rnn_body = recurrent_body
        self.to(device.unwrapped)

    @property
    def state_dim(self) -> Tuple[int, ...]:
        return self.body.input_dim

    @property
    def action_dim(self) -> int:
        return self.actor_head.output_dim

    @property
    def recurrent_body(self) -> RnnBlock:
        return self._rnn_body

    def _features(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tuple[Tensor, RnnState]:
        res = self.body(self.device.tensor(states))
        if rnns is None:
            rnns = self._rnn_body.initial_state(res.size(0), self.device)
        res = self._rnn_body(res, rnns, masks)
        return res

    def policy(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tuple[Policy, RnnState]:
        features, rnn_next = self._features(states, rnns, masks)
        return self.policy_dist(self.actor_head(features)), rnn_next

    def value(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tensor:
        features = self._features(states, rnns, masks)[0]
        return self.critic_head(features).squeeze()

    def forward(
            self,
            states: Union[Array, Tensor],
            rnns: Optional[RnnState] = None,
            masks: Optional[Tensor] = None,
    ) -> Tuple[Policy, Tensor, RnnState]:
        features, rnn_next = self._features(states, rnns, masks)
        policy, value = self.actor_head(features), self.critic_head(features)
        return self.policy_dist(policy), value.squeeze(), rnn_next


def policy_init() -> Initializer:
    """Use small value for policy layer to make policy entroy larger
    """
    return Initializer(weight_init=orthogonal(0.01))


def _make_ac_shared(
        body: NetworkBlock,
        policy_dist: PolicyDist,
        device: Device,
        rnn: Callable[[int, int], RnnBlock],
) -> SharedBodyACNet:
    rnn_ = rnn(body.output_dim, body.output_dim)
    ac_head = LinearHead(body.output_dim, policy_dist.input_dim, policy_init())
    cr_head = LinearHead(body.output_dim, 1)
    return SharedBodyACNet(body, ac_head, cr_head, policy_dist, recurrent_body=rnn_, device=device)


def ac_conv(
        policy: Callable[[int, Device], PolicyDist] = CategoricalDist,
        hidden_channels: Tuple[int, int, int] = (32, 64, 32),
        output_dim: int = 256,
        rnn: Callable[[int, int], RnnBlock] = DummyRnn,
        **kwargs
) -> NetFn:
    """Convolutuion network used for atari experiments
       in A3C paper(https://arxiv.org/abs/1602.01783)
    """
    def _net(state_dim: Tuple[int, int, int], action_dim: int, device: Device) -> SharedBodyACNet:
        body = DqnConv(state_dim, hidden_channels=hidden_channels, output_dim=output_dim, **kwargs)
        policy_dist = policy(action_dim, device)
        return _make_ac_shared(body, policy_dist, device, rnn)
    return _net  # type: ignore


def fc_shared(
        policy: Callable[[int, Device], PolicyDist] = CategoricalDist,
        rnn: Callable[[int, int], RnnBlock] = DummyRnn,
        **kwargs
) -> NetFn:
    """FC body head ActorCritic network
    """
    def _net(state_dim: Tuple[int, ...], action_dim: int, device: Device) -> SharedBodyACNet:
        body = FcBody(state_dim[0], **kwargs)
        policy_dist = policy(action_dim, device)
        return _make_ac_shared(body, policy_dist, device, rnn)
    return _net


def impala_conv(
        policy: Callable[[int, Device], PolicyDist] = CategoricalDist,
        channels: List[int] = [16, 32, 32],
        rnn: Callable[[int, int], RnnBlock] = DummyRnn,
        **kwargs
) -> NetFn:
    """Convolutuion network used in IMPALA
    """
    def _net(state_dim: Tuple[int, int, int], action_dim: int, device: Device) -> SharedBodyACNet:
        body = ResNetBody(state_dim, channels, **kwargs)
        policy_dist = policy(action_dim, device)
        return _make_ac_shared(body, policy_dist, device, rnn)
    return _net  # type: ignore



from abc import ABC, abstractmethod
import torch


class HasStateDict(ABC):
    @abstractmethod
    def state_dict(self) -> dict:
        pass

    def load_state_dict(self, d: dict) -> None:
        for key in d.keys():
            setattr(self, key, d[key])


class TensorStateDict(HasStateDict):
    def state_dict(self) -> dict:
        all_members = self.__dict__.copy()
        for key in all_members.keys():
            value = all_members[key]
            if isinstance(value, torch.Tensor):
                all_members[key] = value.cpu()
        return all_members

    def load_state_dict(self, d: dict) -> None:
        for key in d.keys():
            value = d[key]
            if isinstance(value, torch.Tensor):
                device = getattr(self, key).device
                value = value.to(device)
            setattr(self, key, value)



from abc import ABC, abstractmethod
import copy
import numpy as np
from pathlib import Path
import torch
from torch import nn
from typing import Callable, Generic, Iterable, List, NamedTuple, Optional, Tuple
import warnings
from ..config import Config
from ..lib import mpi
from ..net import DummyRnn, RnnState
from ..envs import EnvExt
from ..prelude import Action, Array, State
from ..utils.log import ExpStats


class EpisodeResult(NamedTuple):
    reward: np.float32
    length: np.int32

    def __repr__(self) -> str:
        return 'Result: reward: {}, length: {}'.format(self.reward, self.length)


class Agent(ABC):
    """Children must call super().__init__(config) first
    """
    def __init__(self, config: Config) -> None:
        self.config = config
        self.logger = config.logger
        self.env = config.env()
        self.total_steps = 0
        self.loss_stat = ExpStats()

    @abstractmethod
    def members_to_save(self) -> Tuple[str, ...]:
        """Here you can specify members you want to save.

    Examples::
        def members_to_save(self):
            return "net", "target"
        """
        pass

    @abstractmethod
    def train_episodes(self, max_steps: int) -> Iterable[List[EpisodeResult]]:
        """Train the agent.
        """
        pass

    @abstractmethod
    def eval_action(self, state: Array) -> Action:
        """Return the best action according to training results.
        """
        pass

    @property
    @abstractmethod
    def update_steps(self) -> int:
        pass

    def eval_reset(self) -> None:
        pass

    def set_mode(self, train: bool = True) -> None:
        pass

    def report_loss(self, **kwargs) -> None:
        if not mpi.IS_MPI_ROOT:
            return
        self.loss_stat.update(kwargs)
        if self.update_steps % self.config.loss_log_freq == 0:
            d = self.loss_stat.report_and_reset()
            d['update-steps'] = self.update_steps
            self.logger.exp('loss', d)

    def close(self) -> None:
        self.env.close()

    def __eval_episode(
            self,
            select_action: Callable[[Array], Action],
            render: bool
    ) -> Tuple[EpisodeResult, EnvExt]:
        total_reward = 0.0
        steps = 0
        env = self.config.eval_env
        if self.config.seed is not None:
            env.seed(self.config.seed)
        state = env.reset()
        while True:
            if render:
                env.render()
            state = env.extract(state)
            action = select_action(state)
            state, reward, done, info = env.step(action)
            steps += 1
            total_reward += reward
            res = self._result(done, info, total_reward, steps)
            if res is not None:
                self.eval_reset()
                return (res, env)

    def _result(
            self,
            done: bool,
            info: dict,
            total_reward: float,
            episode_length: int,
    ) -> Optional[EpisodeResult]:
        if self.env.use_reward_monitor:
            if 'episode' in info:
                return EpisodeResult(info['episode']['r'], info['episode']['l'])
        elif done:
            return EpisodeResult(total_reward, episode_length)
        return None

    def random_episode(self, render: bool = False) -> EpisodeResult:
        def act(_state) -> Action:
            return self.config.eval_env.spec.random_action()
        return self.__eval_episode(act, render)[0]

    def random_and_save(self, fname: str, render: bool = False) -> EpisodeResult:
        def act(_state) -> Action:
            return self.config.eval_env.spec.random_action()
        res, env = self.__eval_episode(act, render)
        env.save_history(fname)
        return res

    def eval_episode(self, render: bool = False) -> EpisodeResult:
        return self.__eval_episode(self.eval_action, render)[0]

    def eval_and_save(self, fname: str, render: bool = False) -> EpisodeResult:
        res, env = self.__eval_episode(self.eval_action, render)
        env.save_history(fname)
        return res

    def save(self, filename: str) -> None:
        if not mpi.IS_MPI_ROOT:
            return None
        save_dict = {}
        for member_str in self.members_to_save():
            value = getattr(self, member_str)
            if isinstance(value, nn.DataParallel):
                save_dict[member_str] = value.module.state_dict()
            elif hasattr(value, 'state_dict'):
                save_dict[member_str] = value.state_dict()
            else:
                save_dict[member_str] = value
        log_dir = self.config.logger.log_dir
        if log_dir is None:
            log_dir = Path('.')
        torch.save(save_dict, log_dir.joinpath(filename))

    def load(self, filename: str) -> None:
        if not mpi.IS_MPI_ROOT:
            return None
        saved_dict = torch.load(filename, map_location=self.config.device.unwrapped)
        #  For backward compatibility, we need to check both index and name
        for idx, member_str in enumerate(self.members_to_save()):
            if idx in saved_dict:
                saved_item = saved_dict[idx]
            elif member_str in saved_dict:
                saved_item = saved_dict[member_str]
            else:
                warnings.warn('Member {} wasn\'t loaded'.format(member_str))
                continue
            mem = getattr(self, member_str)
            if isinstance(mem, nn.DataParallel):
                mem.module.load_state_dict(saved_item)
            elif hasattr(mem, 'state_dict'):
                mem.load_state_dict(saved_item)
            else:
                setattr(self, member_str, saved_item)


class OneStepAgent(Agent, Generic[State]):
    @abstractmethod
    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        pass

    @property
    def update_steps(self) -> int:
        return self.total_steps

    def train_episodes(self, max_steps: int) -> Iterable[List[EpisodeResult]]:
        if self.config.seed is not None:
            self.env.seed(self.config.seed)
        state = self.env.reset()
        total_reward = 0.0
        episode_length = 0
        while True:
            state, reward, done, info = self.step(state)
            self.total_steps += 1
            total_reward += reward
            episode_length += 1
            res = self._result(done, info, total_reward, episode_length)
            if res is not None:
                yield [res]
                state = self.env.reset()
                total_reward = 0.0
                episode_length = 0
            if self.total_steps >= max_steps:
                break


class NStepParallelAgent(Agent, Generic[State]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.rewards = np.zeros(config.nworkers, dtype=np.float32)
        self.episode_length = np.zeros(config.nworkers, dtype=np.int)
        self.episode_results: List[EpisodeResult] = []
        self.penv = config.parallel_env()
        self.eval_rnns: RnnState = DummyRnn.DUMMY_STATE
        self.step_width = self.config.nsteps * self.config.nworkers * mpi.global_size()

    def eval_parallel(
            self,
            n: Optional[int] = None,
            entropy: Optional[Array[float]] = None,
    ) -> List[EpisodeResult]:
        reserved = (
            copy.deepcopy(self.rewards),
            copy.deepcopy(self.episode_length),
            copy.deepcopy(self.episode_results),
        )
        self.rewards.fill(0.0)
        self.episode_length.fill(0)
        self.episode_results.clear()
        if n is None:
            n = self.config.nworkers
        self.config.set_parallel_seeds(self.penv)

        states = self.penv.reset()
        mask = self.config.device.ones(self.config.nworkers)
        while True:
            actions = self.eval_action_parallel(self.penv.extract(states), mask, entropy)
            states, rewards, done, info = self.penv.step(actions)
            self.episode_length += 1
            self.rewards += rewards
            self.report_reward(done, info)
            if n <= len(self.episode_results):
                break

        res = self.episode_results
        self.rewards, self.episode_length, self.episode_results = reserved
        self.eval_reset()
        return res

    @abstractmethod
    def eval_action_parallel(
            self,
            states: Array,
            mask: torch.Tensor,
            ent: Optional[Array[float]] = None,
    ) -> Array[Action]:
        pass

    @abstractmethod
    def nstep(self, states: Array[State]) -> Array[State]:
        pass

    @abstractmethod
    def _reset(self, initial_states: Array[State]) -> None:
        pass

    @property
    def update_steps(self) -> int:
        return self.total_steps // self.step_width

    def rnn_init(self) -> RnnState:
        return DummyRnn.DUMMY_STATE

    def report_reward(self, done: Array[bool], info: Array[dict]) -> None:
        if self.penv.use_reward_monitor:
            for i in filter(lambda i: 'episode' in i, info):
                self.episode_results.append(EpisodeResult(i['episode']['r'], i['episode']['l']))
        else:
            for i in filter(lambda i: done[i], range(self.config.nworkers)):  # type: ignore
                self.episode_results.append(EpisodeResult(self.rewards[i], self.episode_length[i]))
                self.rewards[i] = 0.0
                self.episode_length[i] = 0

    def close(self) -> None:
        self.env.close()
        self.penv.close()

    def train_episodes(self, max_steps: int) -> Iterable[List[EpisodeResult]]:
        self.config.set_parallel_seeds(self.penv)
        states = self.penv.reset()
        self._reset(states)
        while True:
            states = self.nstep(states)
            self.total_steps += self.step_width
            if self.episode_results:
                yield self.episode_results
                self.episode_results = []
            if self.total_steps >= max_steps:
                break



from .a2c import A2cAgent
from .acktr import AcktrAgent
from .aoc import AocAgent
from .base import Agent, EpisodeResult, NStepParallelAgent, OneStepAgent
from .ddqn import DoubleDqnAgent
from .dqn import DqnAgent
from .ppo import PpoAgent




from torch import nn
from torch.optim import Optimizer, RMSprop
from typing import Callable, Dict, List, Optional, Tuple
from .envs import ClassicalControl, DummyParallelEnv, EnvExt, EnvGen, ParallelEnv
from .net import actor_critic, option_critic, value
from .net.prelude import NetFn, Params
from .lib.explore import DummyCooler, Cooler, LinearCooler, Explorer, EpsGreedy
from .lib import mpi
from .lib.kfac import KfacPreConditioner, PreConditioner
from .replay import DqnReplayFeed, ReplayBuffer, UniformReplayBuffer
from .utils import Device, DummyLogger, Logger


class Config:
    def __init__(self) -> None:
        # action/state dims are initialized lazily
        self.action_dim = 0
        self.state_dim: Tuple[int, ...] = (0,)

        # Common parameters
        self.discount_factor = 0.99
        self.device = Device()
        self.grad_clip = 5.0  # I recommend 0.5 for A2C
        self.max_steps = 10000
        self.eval_deterministic = True

        # Replay buffer
        self.replay_batch_size = 10
        self.replay_size = 10000
        self.train_start = 1000
        self.__replay: Callable[[int], ReplayBuffer] = \
            lambda capacity: UniformReplayBuffer(DqnReplayFeed, capacity=capacity)

        # For the cases you can't set seed in constructor, like gym.atari
        self.seed: Optional[int] = None
        self.parallel_seeds: List[int] = []

        # For DQN-like algorithms
        self.sync_freq = 200
        self.__explore: Callable[[], Explorer] = \
            lambda: EpsGreedy(1.0, LinearCooler(1.0, 0.1, 10000))
        self.__eval_explore: Callable[[], Explorer] = lambda: EpsGreedy(0.01, DummyCooler(0.01))

        # For multi worker algorithms
        self.nworkers = 1

        # For n-step algorithms
        self.nsteps = 1

        # For actor-critic algorithms
        self.entropy_weight = 0.01
        self.value_loss_weight = 1.0
        self.use_gae = False
        self.gae_lambda = 1.0
        self.lr_min: Optional[float] = None  # Mujoco: None Atari 0.0

        # For ppo
        self.adv_normalize_eps = 1.0e-5
        self.ppo_minibatch_size = 64  # Mujoco: 64 Atari: 32 * 8
        self.ppo_epochs = 10  # Mujoco: 10 Atari: 3
        self.ppo_clip = 0.2  # Mujoco: 0.2 Atari: 0.1
        self.ppo_value_clip = True
        self.ppo_clip_min: Optional[float] = None  # Mujoco: None Atari: 0.0

        # For option critic
        self.opt_beta_adv_merginal = 0.0
        self.opt_delib_cost = 0.02

        # Logger and logging frequency
        self.logger = Logger() if mpi.IS_MPI_ROOT else DummyLogger()
        self.episode_log_freq = 100
        self.loss_log_freq = 1000
        self.eval_freq = 10000
        self.save_freq = int(1e6)
        self.save_eval_actions = False

        # Optimizer and preconditioner
        self.__optim = lambda params: RMSprop(params, 0.001)
        self.__precond = lambda net: KfacPreConditioner(net)

        # Network
        self.__net: Dict[str, NetFn] = {
            'value': value.fc(),
            'actor-critic': actor_critic.fc_shared(),
            'option-critic': option_critic.fc_shared(num_options=8),
        }

        # Environments
        self.eval_times = 1
        self.__env = lambda: ClassicalControl()
        self.__eval_env: Optional[EnvExt] = None
        self.__paralle_env = lambda env_gen, num_w: DummyParallelEnv(env_gen, num_w)

    def env(self) -> EnvExt:
        env = self.__env()
        if self.state_dim == (0,):
            self.action_dim = env.action_dim
            self.state_dim = env.state_dim
        return env

    def set_env(self, env: EnvGen) -> None:
        self.__env = env

    @property
    def eval_env(self) -> EnvExt:
        if self.__eval_env is None:
            self.__eval_env = self.env()
        return self.__eval_env

    @eval_env.setter
    def eval_env(self, env: EnvExt) -> None:
        if self.state_dim == (0,):
            self.action_dim = env.action_dim
            self.state_dim = env.state_dim
        self.__eval_env = env

    def explorer(self) -> Explorer:
        return self.__explore()

    def set_explorer(self, exp: Callable[[], Explorer]) -> None:
        self.__explore = exp

    def eval_explorer(self) -> Explorer:
        return self.__eval_explore()

    def set_eval_explorer(self, eval_exp: Callable[[], Explorer]) -> None:
        self.__eval_explore = eval_exp

    def optimizer(self, params: Params) -> Optimizer:
        return self.__optim(params)

    def set_optimizer(self, optim: Callable[[Params], Optimizer]) -> None:
        self.__optim = optim

    def preconditioner(self, net: nn.Module) -> PreConditioner:
        return self.__precond(net)

    def set_preconditioner(self, precond: Callable[[nn.Module], PreConditioner]) -> None:
        self.__precond = precond

    def replay_buffer(self) -> ReplayBuffer:
        return self.__replay(self.replay_size)

    def set_replay_buffer(self, replay: Callable[[int], ReplayBuffer]) -> None:
        self.__replay = replay

    def parallel_env(self) -> ParallelEnv:
        penv = self.__parallel_env(self.__env, self.nworkers)
        self.action_dim = penv.action_dim
        self.state_dim = penv.state_dim
        return penv

    def set_parallel_env(self, parallel_env: Callable[[EnvGen, int], ParallelEnv]):
        self.__parallel_env = parallel_env

    def set_parallel_seeds(self, penv: ParallelEnv) -> None:
        if len(self.parallel_seeds) == self.nworkers:
            penv.seed(self.parallel_seeds)
        elif self.seed is not None:
            penv.seed([self.seed] * self.nworkers)

    def net(self, name: str) -> nn.Module:
        assert self.state_dim[0] != 0
        assert self.action_dim != 0
        return self.__net[name](self.state_dim, self.action_dim, self.device)

    def set_net_fn(self, name: str, net: NetFn) -> None:
        self.__net[name] = net

    def _get_cooler(self, initial: float, minimal: Optional[float], update_span: int) -> Cooler:
        if minimal is None:
            return DummyCooler(initial)
        return LinearCooler(initial, minimal, self.max_steps // update_span)

    def lr_cooler(self, initial: float) -> Cooler:
        return self._get_cooler(initial, self.lr_min, self.nsteps * self.nworkers)

    def clip_cooler(self) -> Cooler:
        return self._get_cooler(self.ppo_clip, self.ppo_clip_min, self.nsteps * self.nworkers)

    def __repr__(self) -> str:
        d = filter(lambda t: not t[0].startswith('_Config'), self.__dict__.items())
        return 'Config: ' + str(dict(d))



import numpy as np
import pytest
import torch
from .rms import RunningMeanStd, RunningMeanStdTorch
from ..utils import Device


@pytest.mark.parametrize('shape', [(10,), ()])
def test_rms(shape: tuple) -> None:
    dev = Device()
    dat = torch.randn(1000, dtype=torch.float64, device=dev.unwrapped)
    rms = RunningMeanStd(shape)
    rms_t = RunningMeanStdTorch(torch.Size(shape), device=dev)
    for i in range(10):
        d = dat[i * 100: i * 100 + 100]
        if len(shape) == 1:
            d = d.reshape(10, 10)
        rms.update(d.cpu().numpy())
        rms_t.update(d)
        np.testing.assert_almost_equal(rms.mean, rms_t.mean.detach().cpu().numpy())
        np.testing.assert_almost_equal(rms.var, rms_t.var.detach().cpu().numpy())
        np.testing.assert_almost_equal(rms.std(), rms_t.std().detach().cpu().numpy())


@pytest.mark.parametrize('shape', [(10,), ()])
def test_rmstorch_save(shape: tuple) -> None:
    dev = Device()
    dat = torch.randn(1000, dtype=torch.float64, device=dev.unwrapped)
    rms = RunningMeanStdTorch(torch.Size(shape), device=dev)
    for i in range(10):
        d = dat[i * 100: i * 100 + 100]
        if len(shape) == 1:
            d = d.reshape(10, 10)
        rms.update(d)
    rms_cpu = RunningMeanStdTorch(torch.Size(shape), device=Device(use_cpu=True))
    rms_cpu.load_state_dict(rms.state_dict())
    np.testing.assert_almost_equal(rms.mean.cpu().numpy(), rms_cpu.mean.numpy())



from torch import Tensor
from typing import Any, List, Sequence, Tuple, TypeVar, Union


try:
    from typing import GenericMeta, NamedTupleMeta  # type: ignore

    class GenericNamedMeta(NamedTupleMeta, GenericMeta):
        pass
except ImportError:
    from typing import NamedTupleMeta  # type: ignore
    GenericNamedMeta = NamedTupleMeta  # type: ignore

T = TypeVar('T')
Self = Any
Index = Union[None, int, slice, Tensor, List[Any], Tuple[Any, ...]]


class Array(Sequence[T]):
    @property
    def shape(self) -> tuple:
        ...

    def squeeze(self) -> Self:
        ...

    def transpose(self, *args) -> Self:
        ...

    def mean(self, **kwargs) -> Self:
        ...

    def var(self, **kwargs) -> Self:
        ...

    def __rsub__(self, value: Any) -> Self:
        ...

    def __truediv__(self, rvalue: Any) -> Self:
        ...


Action = TypeVar('Action', int, Array)
State = TypeVar('State')



from copy import deepcopy
import numpy as np
import torch
from torch import nn, Tensor
from typing import Tuple
from .base import OneStepAgent
from ..config import Config
from ..replay import DqnReplayFeed
from ..prelude import Action, Array, State


class DqnAgent(OneStepAgent):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        assert self.env.spec.is_discrete(), 'DQN only supports discrete action spaces'
        self.net = config.net('value')
        self.target_net = deepcopy(self.net)
        self.optimizer = config.optimizer(self.net.parameters())
        self.criterion = nn.MSELoss()
        self.policy = config.explorer()
        self.eval_policy = config.eval_explorer()
        self.replay = config.replay_buffer()
        assert self.replay.feed == DqnReplayFeed
        self.batch_indices = config.device.indices(config.replay_batch_size)

    def set_mode(self, train: bool = True) -> None:
        self.net.train(mode=train)

    def members_to_save(self) -> Tuple[str, ...]:
        return "net", "target_net", "policy", "total_steps"

    @torch.no_grad()
    def eval_action(self, state: Array) -> Action:
        return self.eval_policy.select_action(state, self.net).item()  # type: ignore

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        train_started = self.total_steps > self.config.train_start
        if train_started:
            action = self.policy.select_action(self.env.extract(state), self.net).item()
        else:
            action = self.env.spec.random_action()
        next_state, reward, done, info = self.env.step(action)
        self.replay.append(state, action, reward, next_state, done)
        if train_started:
            self._train()
        return next_state, reward, done, info

    def _q_next(self, next_states: Array) -> Tensor:
        return self.target_net(next_states).max(1)[0]

    def _train(self) -> None:
        obs = self.replay.sample(self.config.replay_batch_size)
        obs = [ob.to_ndarray(self.env.extract) for ob in obs]
        states, actions, rewards, next_states, done = map(np.asarray, zip(*obs))
        with torch.no_grad():
            q_next = self._q_next(next_states)
        q_next *= self.config.device.tensor(1.0 - done) * self.config.discount_factor
        q_next += self.config.device.tensor(rewards)
        q_current = self.net(states)[self.batch_indices, actions]
        loss = self.criterion(q_current, q_next)
        self.optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.net.parameters(), self.config.grad_clip)
        self.optimizer.step()
        self.report_loss(value_loss=loss.item())
        if self.total_steps % self.config.sync_freq == 0:
            self.target_net.load_state_dict(self.net.state_dict())



import gym
from gym.core import Wrapper
import time
from typing import Any, List, Tuple


class RewardMonitor(Wrapper):
    """Based on https://github.com/openai/baselines/blob/master/baselines/bench/monitor.py,
    but modiifed only to report raw rewards before wrapped by 'wrap_deepmind' or else.
    """
    def __init__(self, env: gym.Env) -> None:
        Wrapper.__init__(self, env=env)
        self.tstart = time.time()
        self.rewards: List[float] = []
        self.episode_rewards: List[float] = []
        self.episode_lengths: List[int] = []
        self.episode_times: List[float] = []
        self.total_steps = 0

    def reset(self, **kwargs) -> Any:
        self.rewards = []
        return self.env.reset(**kwargs)

    def step(self, action) -> Tuple[Any, float, bool, dict]:
        ob, rew, done, info = self.env.step(action)
        self.update(ob, rew, done, info)
        return (ob, rew, done, info)

    def update(self, ob, rew, done, info) -> None:
        self.rewards.append(rew)
        self.total_steps += 1
        if not done:
            return
        if not isinstance(info, dict):
            raise ValueError('RewardMonitor assumes info is an instance of dict')
        eprew = sum(self.rewards)
        eplen = len(self.rewards)
        epinfo = {"r": round(eprew, 6), "l": eplen, "t": round(time.time() - self.tstart, 6)}
        self.episode_rewards.append(eprew)
        self.episode_lengths.append(eplen)
        self.episode_times.append(time.time() - self.tstart)
        info['episode'] = epinfo





import io
import os
import re
from setuptools import find_packages, setup


NAME = 'rainy'
AUTHOR = 'Yuji Kanagawa'
EMAIL = 'yuji.kngw.80s.revive@gmail.com'
URL = 'https://github.com/kngwyu/Rainy'
REQUIRES_PYTHON = '>=3.6.0'
DESCRIPTION = 'Algorithm and utilities for deep reinforcement learning'

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'rainy/__init__.py'), 'rt', encoding='utf8') as f:
    VERSION = re.search(r"__version__ = \'(.*?)\'", f.read()).group(1)

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        LONG_DESCRIPTION = '\n' + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION


REQUIRED = [
    'click>=7.0',
    'GitPython>=2.0',
    'gym[atari]>=0.11.0',
    'numpy>=1.15.0',
    'opencv-python>=3.4',
    'Pillow>=5.0',
    'torch>=1.0',
]
EXTRA = {
    'ipython': ['ipython>=7.0', 'matplotlib>=3.0', 'ptpython>=2.0'],
    'bullet': ['pybullet>=2.4'],
    'horovod': ['pybullet>=0.16']
}


setup(
    name=NAME,
    version=VERSION,
    url=URL,
    project_urls={
        'Code': URL,
        'Issue tracker': URL + '/issues',
    },
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIRED,
    extras_require=EXTRA,
    license='Apache2',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)
